from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from functools import reduce
from typing import Any, Callable, List, Tuple, Union

from ...datatypes import EntityType
from ...db import DbChange
from ...utils import group_by, groupby_property
from ...utils.account import get_id_group


class ApplicationStrategy(Enum):
    TX = "transaction"
    BATCH = "batch"

    def __str__(self):
        return str(self.value)


class DeltaUpdate(ABC):
    @abstractmethod
    def merge(self, other_delta):
        pass


@dataclass
class DeltaScalar(DeltaUpdate):
    value: int

    def merge(self, other):
        return DeltaScalar(self.value + other.value)


@dataclass
class DeltaValue(DeltaUpdate):
    value: int
    fiat_values: List[int]

    @classmethod
    def from_db(Cls, db_row):
        return Cls(value=db_row.value, fiat_values=list(db_row.fiat_values))

    def merge(self, other):
        if other is None:
            return self
        assert self.fiat_values is not None and other.fiat_values is not None
        assert len(self.fiat_values) == len(other.fiat_values)
        return DeltaValue(
            value=self.value + other.value,
            fiat_values=[sum(i) for i in zip(self.fiat_values, other.fiat_values)],
        )


def merge_asset_dicts(d1, d2):  # probably better to wrap in class and define __add__
    d = {}
    for k in set(d1.keys()) | set(d2.keys()):
        d[k] = d1.get(k, DeltaValue(0, [0, 0])).merge(d2.get(k, DeltaValue(0, [0, 0])))
    return d


@dataclass
class Tx:
    block_id: int
    tx_id: int
    tx_hash: bytes
    tx_index: int
    failed: bool


@dataclass
class EntityDelta(DeltaUpdate):
    """The identifier is either an address or cluster identifier"""

    identifier: Union[str, int]
    total_received: DeltaValue
    total_spent: DeltaValue
    first_tx_id: int
    last_tx_id: int
    no_incoming_txs: int
    no_outgoing_txs: int

    @classmethod
    def from_db(Cls, db_row, mode: EntityType):
        if mode == EntityType.CLUSTER:
            identifier = db_row.cluster_id
        elif mode == EntityType.ADDRESS:
            identifier = db_row.address
        return Cls(
            identifier=identifier,
            total_received=DeltaValue.from_db(db_row.total_received),
            total_spent=DeltaValue.from_db(db_row.total_spent),
            first_tx_id=db_row.first_tx_id,
            last_tx_id=db_row.last_tx_id,
            no_incoming_txs=db_row.no_incoming_txs,
            no_outgoing_txs=db_row.no_outgoing_txs,
        )

    def merge(self, other_delta):
        assert self.identifier == other_delta.identifier
        return EntityDelta(
            identifier=self.identifier,
            total_received=self.total_received.merge(other_delta.total_received),
            total_spent=self.total_spent.merge(other_delta.total_spent),
            first_tx_id=min(self.first_tx_id, other_delta.first_tx_id),
            last_tx_id=max(self.last_tx_id, other_delta.last_tx_id),
            no_incoming_txs=self.no_incoming_txs + other_delta.no_incoming_txs,
            no_outgoing_txs=self.no_outgoing_txs + other_delta.no_outgoing_txs,
        )


def minusone_respecting_function(x, y, f):
    """
    -1 is a placeholder for first and last tx id in reward traces
    which dont have a tx_id
    """
    if x == -1 and y == -1:
        return -1
    if x == -1:
        return y
    if y == -1:
        return x
    return f(x, y)


@dataclass
class RawEntityTx:
    identifier: Union[str, int]
    is_outgoing: bool
    value: int
    tx_id: int


@dataclass
class RelationDelta(DeltaUpdate):
    src_identifier: Union[str, int]
    dst_identifier: Union[str, int]
    no_transactions: int
    estimated_value: DeltaValue

    @classmethod
    def from_db(Cls, db_row, mode="address"):
        return Cls(
            src_identifier=getattr(db_row, f"src_{mode}"),
            dst_identifier=getattr(db_row, f"dst_{mode}"),
            no_transactions=db_row.no_transactions,
            estimated_value=DeltaValue.from_db(db_row.estimated_value),
        )

    def merge(self, other_delta):
        assert self.src_identifier == other_delta.src_identifier
        assert self.dst_identifier == other_delta.dst_identifier
        return RelationDelta(
            src_identifier=self.src_identifier,
            dst_identifier=self.dst_identifier,
            estimated_value=self.estimated_value.merge(other_delta.estimated_value),
            no_transactions=self.no_transactions + other_delta.no_transactions,
        )


@dataclass
class DbDelta:
    entity_updates: List[EntityDelta]
    new_entity_txs: List[RawEntityTx]
    relation_updates: List[RelationDelta]

    def concat(self, other):
        return DbDelta(
            entity_updates=self.entity_updates + other.entity_updates,
            new_entity_txs=self.new_entity_txs + other.new_entity_txs,
            relation_updates=self.relation_updates + other.relation_updates,
        )

    @staticmethod
    def merge(change_sets: List[DbDelta]) -> "DbDelta":
        return reduce(lambda x, y: x.concat(y), change_sets).compress()

    def to_cluster_delta(self, address_to_cluster_id: Callable[[str], int]):
        eu = deepcopy(self.entity_updates)
        etxs = deepcopy(self.new_entity_txs)
        rel = deepcopy(self.relation_updates)
        for update in eu:
            update.identifier = address_to_cluster_id(update.identifier)

        for update in etxs:
            update.identifier = address_to_cluster_id(update.identifier)

        for update in rel:
            update.src_identifier = address_to_cluster_id(update.src_identifier)
            update.dst_identifier = address_to_cluster_id(update.dst_identifier)

        return DbDelta(
            entity_updates=eu, new_entity_txs=etxs, relation_updates=rel
        ).compress()

    def compress(self):
        grouped = groupby_property(
            self.entity_updates, "identifier", sort_by="first_tx_id"
        )
        entity_updates_merged = {
            k: reduce(lambda x, y: x.merge(y), v) for k, v in grouped.items()
        }
        assert len(entity_updates_merged.keys()) == len(
            set(entity_updates_merged.keys())
        )

        grouped = group_by(
            self.relation_updates, lambda x: (x.src_identifier, x.dst_identifier)
        )
        relations_updates_merged = {
            (src, dst): reduce(lambda x, y: x.merge(y), v)
            for (src, dst), v in grouped.items()
        }

        return DbDelta(
            entity_updates=sorted(
                entity_updates_merged.values(),
                key=lambda x: (x.first_tx_id, x.last_tx_id),
            ),
            new_entity_txs=self.new_entity_txs,
            relation_updates=list(relations_updates_merged.values()),
        )


def prepare_txs_for_ingest(
    delta: List[RawEntityTx],
    resolve_identifier: Callable[[Union[str, int]], int],
    id_bucket_size: int,
    mode: EntityType,
) -> List[DbChange]:
    """
    Creating new address transaction
    """
    changes = []
    for atx in delta:
        ident = resolve_identifier(atx.identifier)

        chng = DbChange.new(
            table=f"{mode}_transactions",
            data={
                f"{mode}_id_group": get_id_group(ident, id_bucket_size),
                f"{mode}_id": ident,
                "tx_id": atx.tx_id,
                "is_outgoing": atx.is_outgoing,
                "value": atx.value,
            },
        )

        changes.append(chng)
    return changes


def prepare_entities_for_ingest(
    delta: List[EntityDelta],
    resolve_identifier: Callable[[Union[str, int]], int],
    resolve_entity: Callable[[Union[str, int]], Any],
    address_to_cluster_id: Callable[[str], int],
    cluster_id_to_address_id: Callable[[int], int],
    new_rel_in: dict,
    new_rel_out: dict,
    id_bucket_size: int,
    get_address_prefix: Callable[[str], Tuple[str, str]],
    mode: EntityType,
) -> Tuple[List[DbChange], int]:
    changes = []
    nr_new_entities = 0
    for update in delta:
        int_ident, entity = (
            resolve_identifier(update.identifier),
            resolve_entity(update.identifier),
        )
        group = get_id_group(int_ident, id_bucket_size)
        if entity is not None:
            """old Address/cluster"""

            assert getattr(entity, f"{mode}_id") == int_ident

            new_value = EntityDelta.from_db(entity, mode=mode).merge(update)

            assert new_value.first_tx_id <= new_value.last_tx_id

            # Nr. of addresses (no_addresses) is currently not updated for clusters
            # Since no merges happen there should not be a difference
            generic_data = {
                "no_incoming_txs": new_value.no_incoming_txs,
                "no_outgoing_txs": new_value.no_outgoing_txs,
                "first_tx_id": new_value.first_tx_id,
                "last_tx_id": new_value.last_tx_id,
                "total_received": new_value.total_received,
                "total_spent": new_value.total_spent,
                "in_degree": entity.in_degree + new_rel_in[update.identifier],
                "out_degree": entity.out_degree + new_rel_out[update.identifier],
                f"{mode}_id": int_ident,
                f"{mode}_id_group": group,
            }

            chng = DbChange.update(
                table=f"{mode}",
                data=generic_data,
            )

            changes.append(chng)
        else:
            """new address/cluster"""
            assert update.first_tx_id <= update.last_tx_id
            nr_new_entities += 1

            data = {
                "no_incoming_txs": update.no_incoming_txs,
                "no_outgoing_txs": update.no_outgoing_txs,
                "first_tx_id": update.first_tx_id,
                "last_tx_id": update.last_tx_id,
                "total_received": update.total_received,
                "total_spent": update.total_spent,
                f"{mode}_id": int_ident,
                f"{mode}_id_group": group,
                "in_degree": new_rel_in[update.identifier],
                "out_degree": new_rel_out[update.identifier],
            }

            if mode == EntityType.ADDRESS:
                data["address"] = update.identifier
                data["cluster_id"] = address_to_cluster_id(update.identifier)
            elif mode == EntityType.CLUSTER:
                data["no_addresses"] = 1

            chng = DbChange.new(table=f"{mode}", data=data)

            changes.append(chng)

            if mode == EntityType.ADDRESS:
                """Clusters don't have a prefix table"""
                address, address_prefix = get_address_prefix(update.identifier)

                changes.append(
                    DbChange.new(
                        table="address_ids_by_address_prefix",
                        data={
                            "address": address,
                            "address_id": int_ident,
                            "address_prefix": address_prefix,
                        },
                    )
                )
            elif mode == EntityType.CLUSTER:
                """Add a entry to cluster addresses"""
                changes.append(
                    DbChange.new(
                        table="cluster_addresses",
                        data={
                            "address_id": cluster_id_to_address_id(int_ident),
                            "cluster_id": int_ident,
                            "cluster_id_group": group,
                        },
                    )
                )

    return changes, nr_new_entities


def prepare_relations_for_ingest(
    delta: List[RelationDelta],
    resolve_identifier: Callable[[Union[str, int]], int],
    inrelations: dict,
    outrelations: dict,
    id_bucket_size: int,
    mode: EntityType,
) -> Tuple[List[DbChange], dict, dict, int]:
    new_relations_in = defaultdict(int)
    new_relations_out = defaultdict(int)
    nr_new_rel = 0

    changes = []
    """ Merging relations deltas """
    for relations_update in delta:
        outr = outrelations[
            (relations_update.src_identifier, relations_update.dst_identifier)
        ].result_or_exc.one()
        inr = inrelations[
            (relations_update.src_identifier, relations_update.dst_identifier)
        ].result_or_exc.one()

        assert (outr is None) == (inr is None)

        id_src = resolve_identifier(relations_update.src_identifier)
        id_dst = resolve_identifier(relations_update.dst_identifier)
        src_group = get_id_group(id_src, id_bucket_size)
        dst_group = get_id_group(id_dst, id_bucket_size)

        if outr is None:
            """new address/cluster relation to insert"""
            new_relations_out[relations_update.src_identifier] += 1
            new_relations_in[relations_update.dst_identifier] += 1

            chng_in = DbChange.new(
                table=f"{mode}_incoming_relations",
                data={
                    f"dst_{mode}_id_group": dst_group,
                    f"dst_{mode}_id": id_dst,
                    f"src_{mode}_id": id_src,
                    "no_transactions": relations_update.no_transactions,
                    "estimated_value": relations_update.estimated_value,
                },
            )
            chng_out = DbChange.new(
                table=f"{mode}_outgoing_relations",
                data={
                    f"src_{mode}_id_group": src_group,
                    f"src_{mode}_id": id_src,
                    f"dst_{mode}_id": id_dst,
                    "no_transactions": relations_update.no_transactions,
                    "estimated_value": relations_update.estimated_value,
                },
            )
            nr_new_rel += 2
        else:
            """update existing adddress relation"""
            nv = DeltaValue.from_db(outr.estimated_value).merge(
                relations_update.estimated_value
            )

            chng_in = DbChange.update(
                table=f"{mode}_incoming_relations",
                data={
                    f"dst_{mode}_id_group": dst_group,
                    f"dst_{mode}_id": id_dst,
                    f"src_{mode}_id": id_src,
                    "no_transactions": outr.no_transactions
                    + relations_update.no_transactions,
                    "estimated_value": nv,
                },
            )

            chng_out = DbChange.update(
                table=f"{mode}_outgoing_relations",
                data={
                    f"src_{mode}_id_group": src_group,
                    f"src_{mode}_id": id_src,
                    f"dst_{mode}_id": id_dst,
                    "no_transactions": outr.no_transactions,
                    "estimated_value": nv,
                },
            )

        changes.append(chng_in)
        changes.append(chng_out)

    return changes, new_relations_in, new_relations_out, nr_new_rel


class Action(Enum):
    CONTINUE = 1
    BREAK = 2



import logging
import time
from typing import List

from ..datatypes import FlowDirection
from ..utils.signals import graceful_ctlc_shutdown
from .abstract import EventNotifier, FlowProvider, WatcherState, WatchpointProvider

logger = logging.getLogger(__name__)


class FlowWatcher:
    def __init__(
        self,
        state: WatcherState,
        watchpoints: WatchpointProvider,
        flow_provider: FlowProvider,
        notifiers: List[EventNotifier],
        new_block_backoff_sec: int = 600,
    ):
        self.provider = flow_provider
        self.notifiers = notifiers
        self.state = state
        self.watchpoints = watchpoints
        self.new_block_backoff_sec = new_block_backoff_sec

    def __enter__(self):
        self.state.__enter__()
        self.provider.__enter__()
        for no in self.notifiers:
            no.__enter__()
        self.watchpoints.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.state.__exit__(exc_type, exc_value, traceback)
        self.provider.__exit__(exc_type, exc_value, traceback)
        for no in self.notifiers:
            no.__exit__(exc_type, exc_value, traceback)
        self.watchpoints.__exit__(exc_type, exc_value, traceback)

    def watch(self):
        try:
            with graceful_ctlc_shutdown() as shutdown_initialized:
                while True:
                    next_block = self.state.get_next_watch_block()

                    flows = self.provider.get_flows_for_block(next_block)

                    if next_block % 1000 == 0:
                        logger.info(f"Done with block {next_block}")

                    if shutdown_initialized():
                        self.state.persist()
                        return

                    if flows is None:
                        wait_sec = int(self.new_block_backoff_sec * 1.3)
                        logger.info(f"No data found for block, waiting {wait_sec}s.")
                        time.sleep(wait_sec)
                        continue

                    for flow, raw_flow in flows:
                        if self.watchpoints.is_watched(flow.address):
                            wconfig = self.watchpoints.get_configuration(flow.address)
                            if (
                                (
                                    wconfig.on_incoming
                                    and flow.direction == FlowDirection.IN
                                )
                                or (
                                    wconfig.on_outgoing
                                    and flow.direction == FlowDirection.OUT
                                )
                            ) and (
                                wconfig.value_gt is None
                                or flow.value > wconfig.value_gt
                            ):
                                for notifier in self.notifiers:
                                    notifier.add_notification(flow, wconfig, raw_flow)

                    for notifier in self.notifiers:
                        notifier.send_notifications()
                    self.state.done_with_block()
        finally:
            self.state.persist()



# flake8: noqa: F401
# from .analytics import AnalyticsDb, KeyspaceConfig
# from .cassandra import CassandraDb, CassandraScope
from .analytics import AnalyticsDb, DbChange
from .cassandra import normalize_cql_statement
from .factory import DbFactory



from typing import List

import grpc

from ...utils import remove_prefix
from .grpc.api.tron_api_pb2 import NumberMessage
from .grpc.api.tron_api_pb2_grpc import WalletStub
from .grpc.core.response_pb2 import TransactionInfoList

# todo check if traces are saved in correct order
# / take note at the correct place that this is unchecked for now


def decode_block_to_traces(block_number: int, block: TransactionInfoList) -> List:
    """decode block of TransactionInfoList protobuf object to get a list of traces

    Args:
        block_number (int): Description
        block (TransactionInfoList): Description

    Returns:
        list: Description
    """
    transactionInfo = block.transactionInfo
    traces_per_block = []

    # i = 73 #  interesting index for block 50_003_457
    # transactionInfo_i = transactionInfo[i]

    trace_index = 0  # unique per block

    for i, transactionInfo_i in enumerate(transactionInfo):
        internal_transactions = transactionInfo_i.internal_transactions

        if len(internal_transactions) == 0:
            continue

        block_number = block_number
        transaction_hash = transactionInfo_i.id.hex()
        internal_index = i

        # convert RepeatedCompositeContainer to list
        for internal_tx in internal_transactions:
            caller_address = (
                internal_tx.caller_address.hex()
            )  # evm style address as str
            transferTo_address = (
                internal_tx.transferTo_address.hex()
            )  # evm style address as str
            callValueInfo = internal_tx.callValueInfo

            note = internal_tx.note.decode("utf-8")
            rejected = internal_tx.rejected

            if len(callValueInfo) == 0:
                call_info_index = None
                call_token_id = None
                call_value = None
                data = {
                    "block_number": block_number,
                    "transaction_hash": transaction_hash,
                    "internal_index": internal_index,
                    "caller_address": caller_address,
                    "transferTo_address": transferTo_address,
                    "call_info_index": call_info_index,
                    "call_token_id": call_token_id,
                    "call_value": call_value,
                    "note": note,
                    "rejected": rejected,
                    "trace_index": trace_index,
                }
                trace_index += 1
                traces_per_block.append(data)
                continue

            for j, callValueInfo_j in enumerate(callValueInfo):
                call_info_index = j
                call_token_id = (
                    callValueInfo_j.tokenId
                )  # this returns an empty string if it is TRX #
                call_token_id = None if call_token_id == "" else int(call_token_id)
                call_value = callValueInfo_j.callValue

                internal_transactions = list(internal_transactions)
                data = {
                    "block_number": block_number,
                    "transaction_hash": transaction_hash,
                    "internal_index": internal_index,
                    "caller_address": caller_address,
                    "transferTo_address": transferTo_address,
                    "call_info_index": call_info_index,
                    "call_token_id": call_token_id,
                    "call_value": call_value,
                    "note": note,
                    "rejected": rejected,
                    "trace_index": trace_index,
                }
                trace_index += 1
                traces_per_block.append(data)

    return traces_per_block


def decode_fees(block_number: int, block: TransactionInfoList) -> List:
    transactionInfo = block.transactionInfo

    return [
        {
            "fee": tx.fee,
            "tx_hash": tx.id.hex(),
            "energy_usage": tx.receipt.energy_usage,
            "energy_fee": tx.receipt.energy_fee,
            "origin_energy_usage": tx.receipt.origin_energy_usage,
            "energy_usage_total": tx.receipt.energy_usage_total,
            "net_usage": tx.receipt.net_usage,
            "net_fee": tx.receipt.net_fee,
            "result": tx.receipt.result,
            "energy_penalty_total": tx.receipt.net_fee,
        }
        for tx in transactionInfo
    ]


class TronExportTracesJob:
    def __init__(
        self,
        start_block: int,
        end_block: int,
        batch_size: int,
        grpc_endpoint: str,
        max_workers: int,
    ):
        self.start_block = start_block
        self.end_block = end_block
        self.batch_size = batch_size
        self.grpc_endpoint = remove_prefix(grpc_endpoint, "grpc://")
        self.max_workers = max_workers

    def run(self):
        channel = grpc.insecure_channel(self.grpc_endpoint)
        wallet_stub = WalletStub(channel)

        traces = []
        fees = []
        for i in range(self.start_block, self.end_block + 1):
            block = wallet_stub.GetTransactionInfoByBlockNum(NumberMessage(num=i))

            traces_per_block = decode_block_to_traces(i, block)
            fees_per_block = decode_fees(i, block)

            traces.extend(traces_per_block)
            fees.extend(fees_per_block)

        return traces, fees



import math
from abc import ABC, abstractmethod
from collections import Counter
from functools import reduce
from typing import Optional, Union

from bitarray import bitarray
from bitarray.util import ba2int
from cashaddress.convert import InvalidAddress as BCHInvalidAddress

from .accountmodel import hex_to_bytes, strip_0x
from .bch import bch_address_to_legacy
from .tron import evm_to_tron_address_string, tron_address_to_evm_string

# from typeguard import typechecked


class InvalidAddress(Exception):
    pass


class BitCoder:
    def __init__(self, alpha: str, bit_width: int):
        assert (
            Counter(alpha).most_common(1)[0][1] == 1
        ), f"Alphabet has duplicate {Counter(alpha).most_common(1)}"
        bits = math.ceil(math.log(len(alpha) + 1, 2))
        assert bit_width == bits, "CAUTION: Bit width change, this breaks decoding"

        codewords = [(k, "{:b}".format(i + 1).zfill(bits)) for i, k in enumerate(alpha)]
        self.decode_table = {code: letter for letter, code in codewords}
        self.encode_table = {letter: bitarray(code) for letter, code in codewords}
        self.bits = bits

    @property
    def alphabet(self) -> str:
        return "".join(self.encode_table.keys())

    def encode_bitarray(self, str_data: str) -> bitarray:
        try:
            return reduce(
                lambda b, char: b + self.encode_table[char], str_data, bitarray()
            )
        except KeyError as e:
            raise InvalidAddress(f"{e} not in alphabet") from e

    def encode(self, str_data: str) -> bytes:
        return self.encode_bitarray(str_data).tobytes()

    def decode_bitarray(self, arr: bitarray) -> str:
        chunks = [arr[i : i + self.bits] for i in range(0, len(arr), self.bits)]
        try:
            return "".join(
                [
                    self.decode_table[code.to01()]
                    for code in chunks
                    if len(code) == self.bits and ba2int(code) != 0
                ]
            )
        except KeyError as e:
            raise InvalidAddress(f"{e} not in alphabet") from e

    def decode(self, byts: bytes) -> str:
        arr = bitarray()
        arr.frombytes(byts)
        return self.decode_bitarray(arr)


class Base58BitCoder(BitCoder):
    def __init__(self):
        super().__init__(
            "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz", bit_width=6
        )


class Bech32BitCoder(BitCoder):
    def __init__(self):
        # bech32 alphabet + b1 to support bc1 prefix mainnet addresses
        super().__init__("qpzry9x8gf2tvdw0s3jn54khce6mua7lb1", bit_width=6)


class AddressConverter(ABC):
    @property
    def supports_partial_address_conversion(self) -> bool:
        return False

    def to_canonical_address_str(self, address: str) -> str:
        return address

    def to_user_address_str(self, address: str) -> str:
        return address

    @abstractmethod
    def to_bytes(self, address: str) -> bytes:
        pass

    @abstractmethod
    def to_str(self, address_byte: bytes) -> str:
        pass


class AddressConverterEth(AddressConverter):
    def to_canonical_address_str(self, address: str) -> str:
        return strip_0x(address.lower())

    def to_bytes(self, address: str) -> bytes:
        try:
            r = hex_to_bytes(self.to_canonical_address_str(address))
        except ValueError as e:
            raise InvalidAddress(f"{address} is not a valid evm address.") from e
        if r is None:
            raise InvalidAddress(f"Could not convert address {address} to bytes.")
        return r

    def to_str(self, address_byte: bytes) -> str:
        return self.to_user_address_str(address_byte.hex())


class AddressConverterTrx(AddressConverterEth):
    def to_canonical_address_str(self, address: str) -> str:
        return strip_0x(tron_address_to_evm_string(address))

    def to_user_address_str(self, address: str):
        return evm_to_tron_address_string(address)


class AddressConverterBtcLike(AddressConverter):
    def __init__(self, bech32_prefix: Optional[str] = None):
        self.base58codec = Base58BitCoder()
        if bech32_prefix is not None:
            self.bech32codec = Bech32BitCoder()
            self.bech32_prefix_bit = self.bech32codec.encode_bitarray(bech32_prefix)
            # make sure that the bech32 prefix is not a valid base58 prefix
            # https://en.bitcoin.it/wiki/List_of_address_prefixes
            assert (
                self.base58codec.decode_bitarray(self.bech32_prefix_bit) == "YCRa"
                or self.base58codec.decode_bitarray(self.bech32_prefix_bit) == "ZRa"
            ), (
                "if new bech prefixes are added make sure that there are no possible "
                "collisions with other addresses e.g. ZRa is the base58 "
                "version of bc1, no valid btc adress starts this way so "
                "no decoding collisions are possible."
            )
        self.bech32_prefix = bech32_prefix

    @property
    def supports_bech32(self) -> bool:
        return self.bech32_prefix is not None

    def get_codec(self, address: Union[str, bytes]) -> BitCoder:
        if self.supports_bech32 and self.is_bech32(address):
            return self.bech32codec
        else:
            return self.base58codec

    def is_bech32(self, address: Union[str, bytes]) -> bool:
        return (
            self.is_bech32_bytes(address)
            if isinstance(address, bytes)
            else self.is_bech32_str(address)
        )

    def is_bech32_str(self, address: str) -> bool:
        return address.startswith(self.bech32_prefix) if self.bech32_prefix else False

    def is_bech32_bytes(self, address: bytes) -> bool:
        arr = bitarray()
        arr.frombytes(address)
        return arr.find(self.bech32_prefix_bit) == 0 if self.bech32_prefix else False

    def to_bytes(self, address: str) -> bytes:
        c = self.get_codec(address)
        return c.encode(self.to_canonical_address_str(address))

    def to_str(self, address_byte: bytes) -> str:
        c = self.get_codec(address_byte)
        return c.decode(address_byte)


class AddressConverterBch(AddressConverterBtcLike):
    def __init__(self):
        super().__init__(bech32_prefix=None)

    def to_canonical_address_str(self, address: str) -> str:
        try:
            return bch_address_to_legacy(address)
        except BCHInvalidAddress as e:
            raise InvalidAddress(f"{address} is not a valid bch address.") from e


converters = {
    "eth": AddressConverterEth(),
    "trx": AddressConverterTrx(),
    "ltc": AddressConverterBtcLike(bech32_prefix="ltc1"),
    "btc": AddressConverterBtcLike(bech32_prefix="bc1"),
    "bch": AddressConverterBch(),
    "zec": AddressConverterBtcLike(bech32_prefix=None),
}


# @typechecked
def address_to_bytes(network: str, address: str) -> bytes:
    c = converters.get(network.lower(), None)
    if c is not None:
        return c.to_bytes(address)
    else:
        raise ValueError("No address converter configured for network {network}")


# @typechecked
def address_to_str(network: str, address: bytes) -> str:
    c = converters.get(network.lower(), None)
    if c is not None:
        return c.to_str(address)
    else:
        raise ValueError("No address converter configured for network {network}")



from typing import Union

from ..utils import hex_to_bytes


class AddressUtxo:
    def __init__(self, adr: Union[str], config):
        """Init an address instance.

        Args:
            adr (Union[str, bytearray]): address
            config (ConfigRow): entry from the config table in the transformed keyspace

        Raises:
            Exception: Description
            ValueError: Description
        """
        self.prefix_length = int(config.address_prefix_length)
        self.bech32_prefix = config.bech_32_prefix
        if type(adr) == str:
            self.address = adr
        else:
            raise Exception("Unknown address format")

    @property
    def is_bech32(self):
        return (
            self.bech32_prefix is not None
            and len(self.bech32_prefix) > 0
            and self.address.startswith(self.bech32_prefix)
        )

    @property
    def prefix(self) -> str:
        if self.is_bech32:
            s = len(self.bech32_prefix)
            return self.db_encoding[s : s + self.prefix_length]
        else:
            return self.db_encoding[: self.prefix_length]

    @property
    def db_encoding(self) -> str:
        return self.address

    @property
    def db_encoding_query(self) -> str:
        return f"'{self.address}'"


class AddressAccount:
    def __init__(self, adr: Union[str, bytearray], config):
        """Init an address instance.

        Args:
            adr (Union[str, bytearray]): address
            config (ConfigRow): entry from the config table in the transformed keyspace

        Raises:
            Exception: Description
            ValueError: Description
        """
        self.prefix_length = int(config.address_prefix_length)
        if type(adr) == str:
            self.address_bytes = hex_to_bytes(adr)
        elif type(adr) == bytearray:
            self.address_bytes = adr
        elif type(adr) == bytes:
            self.address_bytes = bytearray(adr)
        else:
            raise Exception("Unknown address type")

        if len(self.address_bytes) != 20:
            raise ValueError(
                f"Address is not the right length {len(self.address_bytes)}"
            )

    @property
    def hex(self) -> str:  # noqa
        return self.address_bytes.hex()

    @property
    def db_encoding(self) -> str:
        return self.bytearray
        # return f"0x{self.hex}"

    @property
    def db_encoding_query(self) -> str:
        return f"0x{self.hex}"

    @property
    def prefix(self) -> str:
        return self.hex.upper()[: self.prefix_length]

    @property
    def bytearray(self) -> bytearray:  # noqa
        return self.address_bytes


class AddressAccountTrx:
    def __init__(self, adr: Union[str, bytearray], config):
        raise NotImplementedError("AddressAccountTrx not implemented yet")



from dataclasses import asdict, dataclass, fields
from datetime import datetime as dt
from typing import Any, Dict, List, Optional

from ..config import GRAPHSENSE_DEFAULT_DATETIME_FORMAT
from ..db import DbFactory


@dataclass
class DbSummaryRecord:
    query_timestamp: str
    currency: str
    raw_keyspace: str
    raw_highest_block: int
    raw_no_blocks: Optional[int]
    raw_no_txs: Optional[int]
    raw_timestamp: Optional[int]
    transformed_keyspace: str
    transformed_highest_block: int
    transformed_no_blocks: int
    transformed_no_address_relations: int
    transformed_no_addresses: int
    transformed_no_txs: int
    transformed_timestamp: int
    transformed_no_clusters: Optional[int]
    transformed_no_cluster_relations: Optional[int]
    transformed_no_blocks_transform: Optional[int]
    transformed_timestamp_transform: Optional[int]

    @classmethod
    def get_fields(cls) -> List[str]:
        return [field.name for field in fields(cls)]

    def get_dict(self) -> Dict[str, Any]:
        return asdict(self)


def get_db_summary_record(env: str, currency: str) -> DbSummaryRecord:
    with DbFactory().from_config(env, currency) as db:
        raw_stats = db.raw.get_summary_statistics()
        t_stats = db.transformed.get_summary_statistics()

        return DbSummaryRecord(
            query_timestamp=dt.now().strftime(GRAPHSENSE_DEFAULT_DATETIME_FORMAT),
            currency=currency,
            raw_keyspace=db.raw.get_keyspace(),
            raw_highest_block=db.raw.get_highest_block(),
            raw_no_blocks=raw_stats.no_blocks if raw_stats is not None else None,
            raw_no_txs=raw_stats.no_txs if raw_stats is not None else None,
            raw_timestamp=raw_stats.timestamp if raw_stats is not None else None,
            transformed_keyspace=db.transformed.get_keyspace(),
            transformed_highest_block=db.transformed.get_highest_block(),
            transformed_no_blocks=t_stats.no_blocks if t_stats is not None else None,
            transformed_no_address_relations=(
                t_stats.no_address_relations if t_stats is not None else None
            ),
            transformed_no_addresses=(
                t_stats.no_addresses if t_stats is not None else None
            ),
            transformed_no_txs=t_stats.no_transactions if t_stats is not None else None,
            transformed_timestamp=t_stats.timestamp if t_stats is not None else None,
            transformed_no_clusters=getattr(t_stats, "no_clusters", None),
            transformed_no_cluster_relations=getattr(
                t_stats, "no_cluster_relations", None
            ),
            transformed_no_blocks_transform=getattr(
                t_stats, "no_blocks_transform", None
            ),
            transformed_timestamp_transform=getattr(
                t_stats, "timestamp_transform", None
            ),
        )



from pathlib import Path
from urllib.parse import urlparse

import pyarrow as pa
import pyarrow.parquet as pq
from pyarrow.fs import HadoopFileSystem

ACCOUNT_SCHEMA_RAW = {
    "log": pa.schema(
        [
            ("partition", pa.int16()),
            ("block_id_group", pa.int32()),
            ("block_id", pa.int32()),
            ("block_hash", pa.binary(32)),
            ("address", pa.binary(20)),
            ("data", pa.large_binary()),
            ("topics", pa.list_(pa.binary(32))),
            ("topic0", pa.binary(32)),
            ("tx_hash", pa.binary(32)),
            ("log_index", pa.int32()),
            ("transaction_index", pa.int32()),
        ]
    ),
    "trace": pa.schema(
        [
            ("partition", pa.int16()),
            ("block_id_group", pa.int32()),
            ("block_id", pa.int32()),
            ("tx_hash", pa.binary(32)),
            ("transaction_index", pa.int32()),
            ("from_address", pa.binary(20)),
            ("to_address", pa.binary(20)),
            ("value", pa.decimal128(38, 0)),
            ("input", pa.large_binary()),
            ("output", pa.large_binary()),
            ("trace_type", pa.string()),
            ("call_type", pa.string()),
            ("reward_type", pa.string()),
            ("gas", pa.int32()),
            ("gas_used", pa.decimal128(38, 0)),
            ("subtraces", pa.int32()),
            ("trace_address", pa.string()),
            ("error", pa.string()),
            ("status", pa.int16()),
            ("trace_id", pa.string()),
            ("trace_index", pa.int32()),
        ]
    ),
    "block": pa.schema(
        [
            ("partition", pa.int16()),
            ("block_id_group", pa.int32()),
            ("block_id", pa.int32()),
            ("block_hash", pa.binary(32)),
            ("parent_hash", pa.binary(32)),
            ("nonce", pa.binary(8)),
            ("sha3_uncles", pa.binary(32)),
            ("logs_bloom", pa.binary(256)),
            ("transactions_root", pa.binary(32)),
            ("state_root", pa.binary(32)),
            ("receipts_root", pa.binary(32)),
            ("miner", pa.binary(20)),
            ("difficulty", pa.decimal128(38, 0)),
            ("total_difficulty", pa.decimal128(38, 0)),
            ("size", pa.int64()),
            ("extra_data", pa.large_binary()),
            ("gas_limit", pa.int32()),
            ("gas_used", pa.int32()),
            ("base_fee_per_gas", pa.decimal128(38, 0)),
            ("timestamp", pa.int32()),
            ("transaction_count", pa.int32()),
        ]
    ),
    "transaction": pa.schema(
        [
            ("partition", pa.int16()),
            ("tx_hash_prefix", pa.string()),
            ("tx_hash", pa.binary(32)),
            ("nonce", pa.int32()),
            ("block_hash", pa.binary(32)),
            ("block_id_group", pa.int32()),
            ("block_id", pa.int32()),
            ("transaction_index", pa.int32()),
            ("from_address", pa.binary(20)),
            ("to_address", pa.binary(20)),
            ("value", pa.decimal128(38, 0)),
            ("gas", pa.int32()),
            ("gas_price", pa.decimal128(38, 0)),
            ("input", pa.large_binary()),
            ("block_timestamp", pa.int32()),
            ("max_fee_per_gas", pa.decimal128(38, 0)),
            ("max_priority_fee_per_gas", pa.decimal128(38, 0)),
            ("transaction_type", pa.decimal128(38, 0)),
            ("receipt_cumulative_gas_used", pa.decimal128(38, 0)),
            ("receipt_gas_used", pa.decimal128(38, 0)),
            ("receipt_contract_address", pa.binary(20)),
            ("receipt_root", pa.binary(32)),
            ("receipt_status", pa.decimal128(38, 0)),
            ("receipt_effective_gas_price", pa.decimal128(38, 0)),
        ]
    ),
}

SCHEMA_MAPPING = {"account": ACCOUNT_SCHEMA_RAW}


def write_parquet(
    path, table_name, parameters, schema_table, partition_cols=["partition"]
):
    if not parameters:
        return
    table = pa.Table.from_pylist(parameters, schema=schema_table[table_name])

    if path.startswith("hdfs://"):
        o = urlparse(path)
        fs = HadoopFileSystem(o.hostname, o.port)

        pq.write_to_dataset(
            table,
            root_path=o.path / table_name,
            partition_cols=partition_cols,
            existing_data_behavior="overwrite_or_ignore",
            filesystem=fs,
        )
    else:
        path = Path(path)
        if path.exists():
            pq.write_to_dataset(
                table,
                root_path=path / table_name,
                partition_cols=partition_cols,
                existing_data_behavior="overwrite_or_ignore",
            )
        else:
            raise Exception(f"Parquet file path not found: {path}")



import contextlib
import logging
import sys
from typing import Dict

import click
from filelock import FileLock
from filelock import Timeout as LockFileTimeout

from ..cli.common import require_currency, require_environment
from ..config import config, currency_to_schema_type
from ..db import DbFactory
from ..schema import GraphsenseSchemas
from ..utils import subkey_get
from .common import INGEST_SINKS
from .factory import IngestFactory
from .parquet import SCHEMA_MAPPING as PARQUET_SCHEMA_MAPPING

logger = logging.getLogger(__name__)


def create_sink_config(sink: str, network: str, ks_config: Dict):
    schema_type = currency_to_schema_type[network]
    sink_config = ks_config.ingest_config.dict().get("raw_keyspace_file_sinks", None)
    if (sink == "parquet" and schema_type.startswith("account")) or sink == "fs-cache":
        file_sink_dir = subkey_get(sink_config, f"{sink}.directory".split("."))
        if file_sink_dir is None:
            logger.warning(
                f"No {sink} file output directory "
                f"({sink}.directory) is configured for {network}. "
                "Ignoring sink."
            )
            return None

        sc = {"output_directory": file_sink_dir}

        if sink == "parquet":
            sc["schema"] = PARQUET_SCHEMA_MAPPING[schema_type]
        if sink == "fs-cache":
            sc["ignore_tables"] = ["trc10", "configuration"]
            if network == "trx":
                sc["key_by"] = {"fee": "tx_hash", "default": "block_id"}

        return sc
    else:
        return {}


@click.group()
def ingest_cli():
    pass


@ingest_cli.group("ingest")
def ingesting():
    """Ingesting raw cryptocurrency data from nodes into the graphsense database"""
    pass


@ingesting.command("from-node")
@require_environment()
@require_currency(required=True)
@click.option(
    "--sinks",
    type=click.Choice(INGEST_SINKS, case_sensitive=False),
    multiple=True,
    default=["cassandra"],
    help="Where the raw data is written to currently"
    " either cassandra, parquet or both (default: cassandra)",
)
@click.option(
    "--start-block",
    type=int,
    required=False,
    help="start block (default: continue from last ingested block)",
)
@click.option(
    "--end-block",
    type=int,
    required=False,
    help="end block (default: last available block)",
)
@click.option(
    "--batch-size",
    type=int,
    default=10,
    help="number of blocks to export (write) at a time (default: 10)"
    "esp. export to parquet benefits from larger values.",
)
@click.option(
    "--timeout",
    type=int,
    required=False,
    default=3600,
    help="Web3 API timeout in seconds (default: 3600s)",
)
@click.option(
    "--info",
    is_flag=True,
    help="display block information and exit",
)
@click.option(
    "-p",
    "--previous_day",
    is_flag=True,
    help="only ingest blocks up to the previous day, "
    "since currency exchange rates might not be "
    "available for the current day",
)
@click.option(
    "--create-schema",
    is_flag=True,
    help="Create database schema if it does not exist",
)
@click.option(
    "--mode",
    type=click.Choice(
        [
            "legacy",
            "utxo_with_tx_graph",
            "utxo_only_tx_graph",
            "account_traces_only",
            "account_fees_only",
            "trx_update_transactions",
        ],
        case_sensitive=False,
    ),
    help="Importer mode",
    default="legacy",
    multiple=False,
)
@click.option(
    "--version",
    type=int,
    default=1,
    help="Which version of the ingest to use, 1: legacy (sequential), 2:parallel"
    "(default: 1)",
)
@click.option(
    "--no-file-lock",
    is_flag=True,
    help="Do not set file lock to avoid conflicting ingests.",
)
def ingest(
    env,
    currency,
    sinks,
    start_block,
    end_block,
    batch_size,
    timeout,
    info,
    previous_day,
    create_schema,
    mode,
    version,
    no_file_lock,
):
    """Ingests cryptocurrency data form the client/node to the graphsense db
    \f
    Args:
        env (str): Environment to work on
        currency (str): currency to work on
    """
    ks_config = config.get_keyspace_config(env, currency)
    sources = ks_config.ingest_config.all_node_references

    if (
        (
            ks_config.schema_type in ["account", "account_trx"]
            and mode.startswith("utxo_")
        )
        or ks_config.schema_type == "utxo"
        and not mode.startswith("utxo_")
    ):
        logger.error(
            f"Mode {mode} is not available for "
            f"{ks_config.schema_type} type currencies. Exiting."
        )
        sys.exit(11)

    if create_schema:
        GraphsenseSchemas().create_keyspace_if_not_exist(
            env, currency, keyspace_type="raw"
        )

    sink_configs = [(k, create_sink_config(k, currency, ks_config)) for k in sinks]

    with DbFactory().from_config(env, currency) as db:
        try:
            lockfile_name = (
                f"/tmp/{db.raw.get_keyspace()}_{db.transformed.get_keyspace()}.lock"
            )
            if not no_file_lock:
                logger.info(f"Try acquiring lockfile {lockfile_name}")
            with (
                contextlib.nullcontext()
                if no_file_lock
                else FileLock(lockfile_name, timeout=1)
            ):
                IngestFactory().from_config(env, currency, version).ingest(
                    db=db,
                    currency=currency,
                    sources=sources,
                    sink_config={k: v for k, v in sink_configs if v is not None},
                    user_start_block=start_block,
                    user_end_block=end_block,
                    batch_size=batch_size,
                    info=info,
                    previous_day=previous_day,
                    provider_timeout=timeout,
                    mode=mode,
                )
        except LockFileTimeout:
            logger.error(
                f"Lockfile {lockfile_name} could not be acquired. "
                "Is another ingest running? If not delete the lockfile."
            )
            sys.exit(911)


@ingesting.command("to-csv")
@require_environment()
@require_currency(required=True)
@click.option(
    "--start-block",
    type=int,
    required=False,
    help="start block (default: 0)",
)
@click.option(
    "--end-block",
    type=int,
    required=False,
    help="end block (default: last available block)",
)
@click.option(
    "--continue",
    "continue_export",
    is_flag=True,
    help="continue from export from position",
)
@click.option(
    "--batch-size",
    type=int,
    default=10,
    help="number of blocks to export (write) at a time (default: 10)",
)
@click.option(
    "--file-batch-size",
    type=int,
    default=1000,
    help="number of blocks to export to a CSV file (default: 1000)",
)
@click.option(
    "--partition-batch-size",
    type=int,
    default=1_000_000,
    help="number of blocks to export in partition (default: 1_000_000)",
)
@click.option(
    "-p",
    "--previous_day",
    is_flag=True,
    help="only ingest blocks up to the previous day, "
    "since currency exchange rates might not be "
    "available for the current day",
)
@click.option(
    "--info",
    is_flag=True,
    help="display block information and exit",
)
@click.option(
    "--timeout",
    type=int,
    required=False,
    default=3600,
    help="Web3 API timeout in seconds (default: 3600s)",
)
def export_csv(
    env,
    currency,
    start_block,
    end_block,
    continue_export,
    batch_size,
    file_batch_size,
    partition_batch_size,
    previous_day,
    info,
    timeout,
):
    """Exports raw cryptocurrency data to gziped csv files.
    \f
    Args:
        env (str): Environment to work on
        currency (str): currency to work on
    """

    if currency != "eth" and currency != "trx":
        logger.error("Csv export is only supported for eth/trx at the moment.")
        sys.exit(11)

    ks_config = config.get_keyspace_config(env, currency)
    source_node_uri = ks_config.ingest_config.get_first_node_reference()
    csv_directory_config = ks_config.ingest_config.raw_keyspace_file_sinks.get(
        "csv", None
    )

    if csv_directory_config is None:
        logger.error(
            "Please provide an output directory in your "
            "config (raw_keyspace_file_sinks.csv.directory)"
        )
        sys.exit(11)

    csv_directory = csv_directory_config.directory

    from .account import export_csv

    with DbFactory().from_config(env, currency) as db:
        export_csv(
            db=db,
            currency=currency,
            provider_uri=source_node_uri,
            directory=csv_directory,
            user_start_block=start_block,
            user_end_block=end_block,
            continue_export=continue_export,
            batch_size=batch_size,
            file_batch_size=file_batch_size,
            partition_batch_size=partition_batch_size,
            info=info,
            previous_day=previous_day,
            provider_timeout=timeout,
        )



import logging
import signal
from contextlib import contextmanager

logger = logging.getLogger(__name__)

SHUTDOWN = False
SHUTDOWN_SIGNAL_CNT = 0


def shutdown_handler(signum, frame):
    global SHUTDOWN, SHUTDOWN_SIGNAL_CNT
    logger.warning("Received STRG-C, will shutdown when reaching a safe spot.")
    SHUTDOWN = True
    SHUTDOWN_SIGNAL_CNT += 1
    if SHUTDOWN_SIGNAL_CNT > 10:
        logger.warning(
            "Received STRG-C more than 10 times. "
            "Disabling shutdown handler. "
            "The next STRG-C will terminate the process. "
            "Be careful, you are on you own now."
        )
        unregister_ctrlc_shutdown_handler()


def shutdown_initialized():
    return SHUTDOWN


def register_ctrlc_shutdown_handler():
    logger.info("Registered graceful ctrlc handler.")
    signal.signal(signal.SIGINT, shutdown_handler)


def unregister_ctrlc_shutdown_handler():
    logger.info("Unregistered graceful ctrlc handler.")
    signal.signal(signal.SIGINT, signal.SIG_DFL)


@contextmanager
def graceful_ctlc_shutdown(*args, **kwds):
    register_ctrlc_shutdown_handler()
    try:
        yield shutdown_initialized
    finally:
        unregister_ctrlc_shutdown_handler()



import logging
from collections import defaultdict
from datetime import datetime
from typing import Any, Callable, Dict, List, NamedTuple, Tuple

from graphsenselib.db import DbChange
from graphsenselib.deltaupdate.update.abstractupdater import TABLE_NAME_DELTA_HISTORY
from graphsenselib.deltaupdate.update.account.modelsdelta import (
    BalanceDelta,
    EntityDeltaAccount,
    RawEntityTxAccount,
    RelationDeltaAccount,
)
from graphsenselib.deltaupdate.update.generic import DeltaValue, Tx
from graphsenselib.utils import DataObject as MutableNamedTuple
from graphsenselib.utils.account import (
    get_id_group,
    get_id_group_with_secondary_addresstransactions,
    get_id_group_with_secondary_relations,
)
from graphsenselib.utils.logging import LoggerScope

logger = logging.getLogger(__name__)


def prepare_txs_for_ingest(
    delta: List[Tx],
    id_bucket_size: int,
    block_bucket_size: int,
    get_transaction_prefix: Callable[[bytes], Tuple[str, str]],
) -> List[DbChange]:
    changes = []

    for update in delta:
        transaction_id = update.tx_id
        transaction_id_group = get_id_group(transaction_id, id_bucket_size)
        transaction = update.tx_hash

        transaction_prefix = get_transaction_prefix(transaction)[1]

        data = {
            "transaction_id_group": transaction_id_group,
            "transaction_id": transaction_id,
            "transaction": transaction,
        }

        chng = DbChange.new(
            table="transaction_ids_by_transaction_id_group",
            data=data,
        )
        changes.append(chng)

        data = {
            "transaction_prefix": transaction_prefix,
            "transaction": transaction,
            "transaction_id": transaction_id,
        }

        chng = DbChange.new(
            table="transaction_ids_by_transaction_prefix",
            data=data,
        )
        changes.append(chng)

        # get transaction ids

        changes.append(chng)

    changes.extend(
        [
            DbChange.new(
                table="block_transactions",
                data={
                    "block_id_group": get_id_group(tx.block_id, block_bucket_size),
                    "block_id": tx.block_id,
                    "tx_id": tx.tx_id,
                },
            )
            for tx in delta
            if not tx.failed
        ]
    )

    return changes


def prepare_balances_for_ingest(
    delta: List[BalanceDelta], id_bucket_size: int, addr_balances: dict
) -> List[DbChange]:
    changes = []
    for balance in delta:
        addr_id = balance.identifier
        addr_group = get_id_group(addr_id, id_bucket_size)
        balance_update = balance.left_join(
            addr_balances.get(addr_id, BalanceDelta(addr_id, {}))
        )

        for assetname, dv in balance_update.asset_balances.items():
            chng = DbChange.update(
                table="balance",
                data={
                    "address_id_group": addr_group,
                    "address_id": addr_id,
                    "currency": assetname,
                    "balance": dv.value,
                },
            )

            changes.append(chng)

    return changes


def prepare_relations_for_ingest(
    delta: List[RelationDeltaAccount],
    hash_to_id: Dict[str, bytes],
    inrelations: dict,
    outrelations: dict,
    id_bucket_size: int,
    relations_nbuckets: int,
) -> Tuple[List[DbChange], dict, dict]:
    new_relations_in = defaultdict(int)
    new_relations_out = defaultdict(int)

    changes = []

    """ Merging relations deltas """
    for relations_update in delta:
        outr = outrelations[
            (relations_update.src_identifier, relations_update.dst_identifier)
        ].result_or_exc.one()
        inr = inrelations[
            (relations_update.src_identifier, relations_update.dst_identifier)
        ].result_or_exc.one()

        # assert (outr is None) == (inr is None)

        id_src = hash_to_id[relations_update.src_identifier]
        id_dst = hash_to_id[relations_update.dst_identifier]

        src_group, src_secondary = get_id_group_with_secondary_relations(
            id_src, id_dst, id_bucket_size, relations_nbuckets
        )
        dst_group, dst_secondary = get_id_group_with_secondary_relations(
            id_dst, id_src, id_bucket_size, relations_nbuckets
        )

        # checking and logging inconsistencies
        if (outr is None) == (inr is None):
            pass
        else:
            debug_msg = "inoutcheck: inconsistency "
            debug_msg += f"src: {bytes.hex(relations_update.src_identifier)} "
            debug_msg += f"dst: {bytes.hex(relations_update.dst_identifier)} "
            debug_msg += f"inr: {inr} "
            debug_msg += f"outr: {outr} "
            logger.warning(debug_msg)

        # missing record patching
        if (outr is None) and (inr is not None):
            outr = MutableNamedTuple(**inr._asdict())

            outr.src_address_id_group = src_group
            outr.src_address_id_secondary_group = src_secondary

        if (inr is None) and (outr is not None):
            inr = MutableNamedTuple(**outr._asdict())

            inr.dst_address_id_group = dst_group
            inr.dst_address_id_secondary_group = dst_secondary

        if outr is None:
            """new address relation to insert"""
            new_relations_out[relations_update.src_identifier] += 1
            new_relations_in[relations_update.dst_identifier] += 1

            chng_in = DbChange.new(
                table="address_incoming_relations",
                data={
                    "dst_address_id_group": dst_group,
                    "dst_address_id_secondary_group": dst_secondary,
                    "dst_address_id": id_dst,
                    "src_address_id": id_src,
                    "no_transactions": relations_update.no_transactions,
                    "value": relations_update.value,
                    "token_values": relations_update.token_values,
                },
            )
            chng_out = DbChange.new(
                table="address_outgoing_relations",
                data={
                    "src_address_id_group": src_group,
                    "src_address_id_secondary_group": src_secondary,
                    "src_address_id": id_src,
                    "dst_address_id": id_dst,
                    "no_transactions": relations_update.no_transactions,
                    "value": relations_update.value,
                    "token_values": relations_update.token_values,
                },
            )

        else:
            """update existing adddress relation"""
            nv = DeltaValue.from_db(outr.value).merge(relations_update.value)

            nv_token = outr.token_values
            nv_token = nv_token if nv_token is not None else {}
            new_token = relations_update.token_values
            keys = set(nv_token.keys()).union(new_token.keys())
            for key in keys:
                if key in nv_token and key in relations_update.token_values:
                    nv_token[key] = DeltaValue.from_db(nv_token[key]).merge(
                        relations_update.token_values[key]
                    )
                elif key in nv_token:
                    pass
                elif key in relations_update.token_values:
                    nv_token[key] = relations_update.token_values[key]

            # assert outr.no_transactions == inr.no_transactions

            chng_in = DbChange.update(
                table="address_incoming_relations",
                data={
                    "dst_address_id_group": dst_group,
                    "dst_address_id_secondary_group": dst_secondary,
                    "dst_address_id": id_dst,
                    "src_address_id": id_src,
                    "no_transactions": outr.no_transactions
                    + relations_update.no_transactions,
                    # outr and and inr should be the same
                    "value": nv,
                    "token_values": nv_token,
                },
            )

            chng_out = DbChange.update(
                table="address_outgoing_relations",
                data={
                    "src_address_id_group": src_group,
                    "src_address_id_secondary_group": src_secondary,
                    "src_address_id": id_src,
                    "dst_address_id": id_dst,
                    "no_transactions": outr.no_transactions
                    + relations_update.no_transactions,
                    "value": nv,
                    "token_values": nv_token,
                },
            )

        changes.append(chng_in)
        changes.append(chng_out)

    return changes, new_relations_in, new_relations_out


def prepare_entities_for_ingest(
    delta: List[EntityDeltaAccount],
    resolve_identifier: Dict[str, int],
    bytes_to_row_address: Dict[str, Any],
    new_rel_in: dict,
    new_rel_out: dict,
    id_bucket_size: int,
    get_address_prefix: Callable[[str], Tuple[str, str]],
) -> Tuple[List[DbChange], int]:
    changes = []
    nr_new_entities = 0
    for update in delta:
        int_ident, entity = (
            resolve_identifier[update.identifier],
            bytes_to_row_address[update.identifier],
        )

        group = get_id_group(int_ident, id_bucket_size)
        if entity is not None:
            """old Address"""

            assert getattr(entity, "address_id") == int_ident

            # recast so we can calculate without handling None all the time
            new_value = EntityDeltaAccount.from_db(entity).merge(update)
            # bytes to hex
            bytes_ = new_value.identifier
            bytes_.hex()
            assert new_value.first_tx_id <= new_value.last_tx_id

            # Nr. of addresses (no_addresses) is currently not updated for clusters
            # Since no merges happen there should not be a difference
            generic_data = {
                "no_incoming_txs": new_value.no_incoming_txs,
                "no_outgoing_txs": new_value.no_outgoing_txs,
                "no_incoming_txs_zero_value": new_value.no_incoming_txs_zero_value,
                "no_outgoing_txs_zero_value": new_value.no_outgoing_txs_zero_value,
                "first_tx_id": new_value.first_tx_id,
                "last_tx_id": new_value.last_tx_id,
                "total_received": new_value.total_received,
                "total_spent": new_value.total_spent,
                "total_tokens_received": new_value.total_tokens_received,
                "total_tokens_spent": new_value.total_tokens_spent,
                "in_degree": entity.in_degree + new_rel_in[update.identifier],
                "out_degree": entity.out_degree + new_rel_out[update.identifier],
                "in_degree_zero_value": entity.in_degree_zero_value,  # too broad
                "out_degree_zero_value": entity.out_degree_zero_value,  # too broad
                "address_id": int_ident,
                "address_id_group": group,
            }

            chng = DbChange.update(
                table="address",
                data=generic_data,
            )

            changes.append(chng)
        else:
            """new address"""
            assert update.first_tx_id <= update.last_tx_id
            nr_new_entities += 1

            data = {
                "no_incoming_txs": update.no_incoming_txs,
                "no_outgoing_txs": update.no_outgoing_txs,
                "no_incoming_txs_zero_value": update.no_incoming_txs_zero_value,
                "no_outgoing_txs_zero_value": update.no_outgoing_txs_zero_value,
                "first_tx_id": update.first_tx_id,
                "last_tx_id": update.last_tx_id,
                "total_received": update.total_received,
                "total_spent": update.total_spent,
                "total_tokens_received": update.total_tokens_received,
                "total_tokens_spent": update.total_tokens_spent,
                "address_id": int_ident,
                "address_id_group": group,
                "in_degree": new_rel_in[update.identifier],
                "out_degree": new_rel_out[update.identifier],
                "in_degree_zero_value": 0,
                # update.no_incoming_txs_zero_value, # too broad
                "out_degree_zero_value": 0,
                # update.no_outgoing_txs_zero_value, #  too broad
                "is_contract": False,
            }
            data["address"] = update.identifier
            chng = DbChange.new(table="address", data=data)
            changes.append(chng)
            address, address_prefix = get_address_prefix(update.identifier)

            changes.append(
                DbChange.new(
                    table="address_ids_by_address_prefix",
                    data={
                        "address": address,
                        "address_id": int_ident,
                        "address_prefix": address_prefix,
                    },
                )
            )
    return changes, nr_new_entities


def prepare_entity_txs_for_ingest(
    delta: List[RawEntityTxAccount],
    id_bucket_size: int,
    currency: str,
    block_bucket_size_address_txs: int,
) -> List[DbChange]:
    """
    Creating new address transaction
    """
    changes = []
    for atx in delta:
        ident = atx.identifier
        is_token_transfer = len(atx.token_values.keys()) > 0
        for tokenname in atx.token_values.keys():
            (
                address_id_group,
                address_id_secondary_group,
            ) = get_id_group_with_secondary_addresstransactions(
                ident, id_bucket_size, atx.block_id, block_bucket_size_address_txs
            )
            chng = DbChange.new(
                table="address_transactions",
                data={
                    "address_id_group": address_id_group,
                    "address_id_secondary_group": address_id_secondary_group,
                    "address_id": ident,
                    "currency": tokenname,
                    "transaction_id": atx.tx_id,
                    "is_outgoing": atx.is_outgoing,
                    "tx_reference": atx.tx_reference,
                },
            )
            changes.append(chng)

        if not is_token_transfer:
            (
                address_id_group,
                address_id_secondary_group,
            ) = get_id_group_with_secondary_addresstransactions(
                ident, id_bucket_size, atx.block_id, block_bucket_size_address_txs
            )
            chng = DbChange.new(
                table="address_transactions",
                data={
                    "address_id_group": address_id_group,
                    "address_id_secondary_group": address_id_secondary_group,
                    "address_id": ident,
                    "currency": currency,
                    "transaction_id": atx.tx_id,
                    "is_outgoing": atx.is_outgoing,
                    "tx_reference": atx.tx_reference,
                },
            )

            changes.append(chng)
    return changes


def get_bookkeeping_changes(
    base_statistics: MutableNamedTuple,
    current_statistics: NamedTuple,
    last_block_processed: int,
    nr_new_address_relations: int,
    nr_new_addresses: int,
    nr_new_tx: int,
    highest_address_id: int,
    runtime_seconds: int,
    bts: Dict[int, datetime],
    new_blocks: int,
    patch_mode: bool,
) -> List[DbChange]:
    """Creates changes for the bookkeeping tables like summary statistics after
    other data has been updated.

    Args:
        base_statistics (MutableNamedTuple): statistics db row, all the other
        parameters are note data is updated in this process
        current_statistics (NamedTuple): Current value of db statistics for comparison
        last_block_processed (int): Last block processed
        nr_new_address_relations (int): Delta new addresses relations in changeset
        nr_new_addresses (int): Delta new addresses in changeset
        nr_new_tx (int): Delta new txs in changeset
        highest_address_id (int): current highest address_id
        runtime_seconds (int): runtime to create the last changes in seconds
        bts (Dict[int, datetime]): mapping from block to its timestamp
        delta values
    """
    changes = []
    with LoggerScope.debug(logger, "Creating summary_statistics updates") as lg:
        lb_date = bts[last_block_processed]
        stats = base_statistics
        stats.no_blocks = current_statistics.no_blocks + new_blocks
        stats.timestamp = int(lb_date.timestamp())
        stats.no_address_relations += nr_new_address_relations
        stats.no_addresses += nr_new_addresses
        stats.no_transactions += nr_new_tx

        statistics = stats.as_dict()

        if current_statistics.no_blocks != stats.no_blocks:
            if not patch_mode:
                assert current_statistics.no_blocks < stats.no_blocks

        changes.append(DbChange.new(table="summary_statistics", data=statistics))

        lg.debug(f"Statistics: {statistics}")

        data_history = {
            "last_synced_block": last_block_processed,
            "last_synced_block_timestamp": lb_date,
            "highest_address_id": highest_address_id,
            "timestamp": datetime.now(),
            "write_new": False,
            "write_dirty": False,
            "runtime_seconds": runtime_seconds,
        }
        changes.append(DbChange.new(table=TABLE_NAME_DELTA_HISTORY, data=data_history))

        lg.debug(f"History: {data_history}")

    return changes



import logging
import sys

import click
from filelock import FileLock
from filelock import Timeout as LockFileTimeout

from ..cli.common import require_currency, require_environment
from .factory import FlowWatcherFactory

logger = logging.getLogger(__name__)


@click.group()
def watch_cli():
    pass


@watch_cli.group()
def watch():
    """Commands for permanently watching cryptocurrency events."""
    pass


@watch.command("money-flows")
@require_environment()
@require_currency(required=True)
@click.option(
    "--state-file",
    type=str,
    required=True,
    help="File to store and read the current state of the watcher, "
    " to avoid dupicate notifications",
)
@click.option(
    "--watchpoints-file",
    type=str,
    required=True,
    help="File that defines the watched addresses.",
)
def watchflows(env, currency, state_file, watchpoints_file):
    """Watches for movements money flows and generates notifications form that.
    \f
    Args:
        env (str): Environment to work on
        currency (str): currency to work on.
    """
    lockfile_name = f"/tmp/gscli_{env}_{currency}_watcher.lock"
    logger.info(f"Try acquiring lockfile {lockfile_name}")
    try:
        with FileLock(lockfile_name, timeout=1):
            with FlowWatcherFactory().file_based_from_config(
                env, currency, state_file, watchpoints_file
            ) as watcher:
                watcher.watch()
    except LockFileTimeout:
        logger.error(
            f"Lockfile {lockfile_name} could not be acquired. "
            "Is another watcher running for the environment?"
            " If not delete the lockfile."
        )
        sys.exit(911)



import hashlib

import base58

from .accountmodel import bytes_to_hex, hex_str_to_bytes, strip_0x

TRON_ADDRESS_PREFIX = b"\x41"


def decode_note(note_hex_str: str) -> str:
    return bytes.fromhex(note_hex_str).decode("utf-8")


def sha256(bts):
    m = hashlib.sha256()
    m.update(bts)
    return m.digest()


def get_tron_address_checksum(addr_bytes_with_prefix: bytes):
    h0 = sha256(addr_bytes_with_prefix)
    h1 = sha256(h0)
    checkSum = h1[0:4]
    return checkSum


def add_tron_prefix(address_bytes, prefix: bytes = TRON_ADDRESS_PREFIX):
    if len(address_bytes) == 20:
        return prefix + address_bytes
    return address_bytes


def strip_tron_prefix(address_bytes, prefix: bytes = TRON_ADDRESS_PREFIX):
    if len(address_bytes) == 21 and address_bytes.startswith(prefix):
        return address_bytes[len(prefix) :]
    return address_bytes


def evm_to_bytes(evm_address_hex: str, prefix: bytes = TRON_ADDRESS_PREFIX) -> bytes:
    return strip_tron_prefix(hex_str_to_bytes(strip_0x(evm_address_hex)), prefix)


def evm_to_tron_address(
    evm_address_hex: str, prefix: bytes = TRON_ADDRESS_PREFIX
) -> bytes:
    # inspired by
    # https://github.com/tronprotocol/tronweb/blob/d8c0d48847c0a2dd1c92f4a93f1e01b31c33dc94/src/utils/crypto.js#L14
    a = add_tron_prefix(hex_str_to_bytes(strip_0x(evm_address_hex)), prefix)
    checkSum = get_tron_address_checksum(a)
    taddress = a + checkSum
    return base58.b58encode(taddress)


def evm_to_tron_address_string(
    evm_address_hex: str, prefix: bytes = TRON_ADDRESS_PREFIX
) -> str:
    return evm_to_tron_address(evm_address_hex, prefix).decode("utf-8")


def tron_address_to_bytes(taddress_str: str, validate: bool = True) -> bytes:
    return tron_address_to_evm(taddress_str, validate)


def tron_address_to_evm(taddress_str: str, validate: bool = True) -> bytes:
    ab = base58.b58decode(taddress_str)
    checkSum = ab[-4:]
    a = ab[:-4]

    # recompute checksum
    checkSumComputed = get_tron_address_checksum(a) if validate else None

    if not validate or all(a == b for a, b in zip(checkSum, checkSumComputed)):
        return strip_tron_prefix(a)
    else:
        raise ValueError(f"Invalid checksum on address {taddress_str}")


def tron_address_to_evm_string(taddress_str: str, validate: bool = True) -> str:
    return "0x" + bytes_to_hex(tron_address_to_evm(taddress_str, validate))


def tron_address_to_legacy(taddress_str: str, validate: bool = True) -> bytes:
    """Converts a tron address to its legacy evm/eth format

    Args:
        taddress_str (str): string of tron base58 encoded address
        validate (bool, optional): if checksum should be checked
    """
    return tron_address_to_evm(taddress_str, validate)


def tron_address_to_legacy_string(taddress_str: str, validate: bool = True) -> str:
    """Converts a tron address to its legacy evm/eth format

    Args:
        taddress_str (str): string of tron base58 encoded address
        validate (bool, optional): if checksum should be checked
    """
    return tron_address_to_evm_string(taddress_str, validate)


def equal_evm_tron(t_address: str, evm_address: str) -> bool:
    try:
        return evm_to_bytes(evm_address) == tron_address_to_bytes(t_address)
    except ValueError:
        return False


def equal_evm_evm(evm_address1: str, evm_address2: str) -> bool:
    try:
        return evm_to_bytes(evm_address1) == evm_to_bytes(evm_address2)
    except ValueError:
        return False


def equal_tron_tron(evm_address1: str, evm_address2: str) -> bool:
    try:
        return tron_address_to_bytes(evm_address1) == tron_address_to_bytes(
            evm_address2
        )
    except ValueError:
        return False


def tron_address_equal(address1: str, address2: str) -> bool:
    return (
        equal_evm_tron(address1, address2)
        or equal_evm_tron(address2, address1)
        or equal_evm_evm(address1, address2)
        or equal_tron_tron(address1, address2)
    )


def partial_tron_to_partial_evm(
    partial_taddress_str: str, prefix: bytes = TRON_ADDRESS_PREFIX
) -> str:
    len_taddress = len(partial_taddress_str)
    len_full_taddress = 34
    padding_length = len_full_taddress - len_taddress

    try:
        partial_ab = base58.b58decode(partial_taddress_str + "u" * padding_length)
        partial_a = partial_ab[: len(partial_ab) - 4]
        partial_evm = strip_tron_prefix(partial_a, prefix)
        if (
            len_taddress == len_full_taddress
        ):  # if it is a full address, don't return only a prefix
            return "0x" + bytes_to_hex(partial_evm)
        return ("0x" + bytes_to_hex(partial_evm))[:len_taddress]
    except Exception as e:
        raise ValueError(f"Could not convert partial TRON address: {e}")



import click
from rich.traceback import install

from .. import __version__
from ..config.cli import config_cli
from ..convert.cli import convert_cli
from ..db.cli import db_cli
from ..deltaupdate.cli import deltaupdate_cli
from ..ingest.cli import ingest_cli
from ..monitoring.cli import monitoring_cli
from ..rates.cli import rates_cli
from ..schema.cli import schema_cli
from ..utils.console import console
from ..utils.logging import configure_logging
from ..utils.slack import ClickSlackErrorNotificationContext
from ..watch.cli import watch_cli
from .common import try_load_config

__author__ = "iknaio"
__copyright__ = "iknaio"
__license__ = "MIT"


@click.group()
def version():
    """Print version info."""
    pass


@version.command("version")
def version_cmd():
    """Display the current version."""
    console.print(__version__)


@click.command(
    cls=click.CommandCollection,
    sources=[
        rates_cli,
        schema_cli,
        ingest_cli,
        db_cli,
        deltaupdate_cli,
        config_cli,
        convert_cli,
        monitoring_cli,
        watch_cli,
        version,
    ],
    epilog="GraphSense - https://graphsense.github.io/",
)
@click.option(
    "-v", "--verbose", count=True, help="One v for warning, two for info etc."
)
@click.option(
    "--config-file",
    type=str,
    help="Change the config file to use. If blank default config location is loaded.",
    required=False,
)
@click.pass_context
def cli(ctx, verbose: int, config_file: str):
    """Commandline interface of graphsense-lib

    graphsense-cli exposes many tools and features to manager your
    graphsense crypto-analytics database.
    \f
    Args:
        verbose (int): One v stands for loglevel warning, two for info and so on...
    """
    config = try_load_config(config_file)
    ctx.with_resource(
        ClickSlackErrorNotificationContext(
            config.get_slack_exception_notification_hook_urls()
        )
    )
    configure_logging(verbose)


def main():
    """install rich as traceback handler for all cli commands"""
    install(show_locals=True, suppress=[click])

    cli()


if __name__ == "__main__":
    main()



# flake8: noqa: E501
import json
import re

from graphsenselib.utils import dict_to_dataobject, dict_with_snake_keys


def preprocess_inputs(text: str, drop=[]):
    tnl = ",\n".join([line for line in text.strip().split("\n")])
    items = json.loads(f"[{tnl}]")
    for item in items:
        keys_to_drop = [k for k in item.keys() if any([re.match(dp, k) for dp in drop])]
        for k in keys_to_drop:
            item.pop(k)
    return [dict_with_snake_keys(item) for item in items]


# These are copies for the tests in the graphsense-transform repo
# src/test/resources/test_txs.json
txs = preprocess_inputs(
    """
{"txIdGroup": 0, "txId": 1, "txHash": "ab100004ba5dd70a143ca2359693c19fc879bb86b259870eed0de573f51d34e6", "coinbase": true, "coinjoin": false, "blockId": 1, "inputs": [], "outputs": [{"address": ["138cWsiAGpW9yqfjMVCCsFcnaiSHyoWMnJ"], "value": 2500000, "txType": 3}], "timestamp": 1230948000, "totalInput": 0, "totalOutput": 2500000}
{"txIdGroup": 0, "txId": 2, "txHash": "ab200013f626405ddebf1a7b2e0af34253d09e80f9ef7f981ec1ec59d6200c1f", "coinbase": true, "coinjoin": false, "blockId": 2, "inputs": [], "outputs": [{"address": ["1HxbaBMF2jXBVnagoHvaA6rLxmaYL8gb8T"], "value": 2500000, "txType": 3}], "timestamp": 1231466400, "totalInput": 0, "totalOutput": 2500000}
{"txIdGroup": 0, "txId": 3, "txHash": "ab20101c4ca509cd1bbb31bef15b96cc5f987e6438f770e105e41d012057930a", "coinbase": false, "coinjoin": false, "blockId": 2, "inputs": [{"address": ["138cWsiAGpW9yqfjMVCCsFcnaiSHyoWMnJ"], "value": 2500000, "txType": 5}], "outputs": [{"address": ["1FDCgJ8m2xDyVmYuankk13XReVC2Zvs5cz"], "value": 1480000, "txType": 5}, {"address": ["1FAkhqm95YnV5Mi7Q5j2Wb8CkbK7Z9zpyB"], "value": 495000, "txType": 5}, {"address": ["3Kawbc5pkpQPfmaoGCBiaw5t2uK6WBoVVe"], "value": 495000, "txType": 5}], "timestamp": 1231466400, "totalInput": 2500000, "totalOutput": 2470000}
{"txIdGroup": 0, "txId": 4, "txHash": "ab300013f626405ddebf1a7b2e0af34253d09e80f9ef7f981ec1ec59d6200c1f", "coinbase": true, "coinjoin": false, "blockId": 3, "inputs": [], "outputs": [{"address": ["12751KvnaTTPMJbHbbercdTV48kL81BjX8"], "value": 2500000, "txType": 3}], "timestamp": 1231552800, "totalInput": 0, "totalOutput": 2500000}
{"txIdGroup": 0, "txId": 5, "txHash": "ab30109910f327700e0f199972eed8ea7c6b1920e965f9cb48a92973e7325046", "coinbase": false, "coinjoin": true, "blockId": 3, "inputs": [{"address": ["1FDCgJ8m2xDyVmYuankk13XReVC2Zvs5cz"], "value": 1480000, "txType": 5}, {"address": ["1HxbaBMF2jXBVnagoHvaA6rLxmaYL8gb8T"], "value": 2500000, "txType": 3}], "outputs": [{"address": ["1E5UPs8bXBq7v7D2b7BYNU19ZviX2LMgEe", "1KLWBUTPXtRB7wvTYq9TYdeo93fAKvWPKs"], "value": 150000, "txType": 6}, {"address": ["1231PgW8KbpwKkvACPhp13fcL6fM5sKGvy"], "value": 1325000, "txType": 3}, {"address": ["1Fufjpf9RM2aQsGedhSpbSCGRHrmLMJ7yY"], "value": 100000, "txType": 3}, {"address": ["1iYSYHTpr2wMShaXTTNUzMohkpuV5p5ep"], "value": 2350000, "txType": 3}], "timestamp": 1231552800, "totalInput": 3980000, "totalOutput": 3925000}
{"txIdGroup": 0, "txId": 6, "txHash": "ab3020c34bed7a86eef78534d72fd38c3c247c25d58be632201cb2027c9d2197", "coinbase": false, "coinjoin": false, "blockId": 3, "inputs": [{"address": ["1FAkhqm95YnV5Mi7Q5j2Wb8CkbK7Z9zpyB"], "value": 495000, "txType": 5}], "outputs": [{"address": ["3Fkx2TFdcHoab4xGgSjhAVh5YBPvbBWjNL"], "value": 140000, "txType": 5}, {"address": ["1FAkhqm95YnV5Mi7Q5j2Wb8CkbK7Z9zpyB"], "value": 345000, "txType": 3}], "timestamp": 1231552800, "totalInput": 495000, "totalOutput": 485000}
{"txIdGroup": 0, "txId": 7, "txHash": "ab4006c39b99683ac8f456721b270786c627ecb246700888315991877024b983", "coinbase": true, "coinjoin": false, "blockId": 4, "inputs": [], "outputs": [{"address": ["1CjPR7Z5ZSyWk6WtXvSFgkptmpoi4UM9BC"], "value": 2500000, "txType": 3}], "timestamp": 1231639200, "totalInput": 0, "totalOutput": 2500000}
{"txIdGroup": 0, "txId": 8, "txHash": "ab40158fdb719f333927a92f72727b996f1ebf565fce2cc83aea04b0f6902c6f", "coinbase": false, "coinjoin": false, "blockId": 4, "inputs": [{"address": ["1E5UPs8bXBq7v7D2b7BYNU19ZviX2LMgEe", "1KLWBUTPXtRB7wvTYq9TYdeo93fAKvWPKs"], "value": 1500000, "txType": 6}, {"address": ["3Kawbc5pkpQPfmaoGCBiaw5t2uK6WBoVVe"], "value": 495000, "txType": 5}, {"address": ["12751KvnaTTPMJbHbbercdTV48kL81BjX8"], "value": 2500000, "txType": 3}], "outputs": [{"address": ["1AGMAXWELLayCyS1vkLXEszESHEcB3LWqa"], "value": 4300000, "txType": 3}, {"address": ["3JX79i9xSSmLEDZ8WrUFsy3WqrpvPoQmPv"], "value": 149000, "txType": 5}], "timestamp": 1231639200, "totalInput": 4495000, "totalOutput": 4449000}
{"txIdGroup": 0, "txId": 9, "txHash": "ab40231bd240c232eb8a8ed1b2267fe9f76d46c892415ab74427195cce5c6cff", "coinbase": false, "coinjoin": false, "blockId": 4, "inputs": [{"address": ["1FAkhqm95YnV5Mi7Q5j2Wb8CkbK7Z9zpyB"], "value": 345000, "txType": 3}], "outputs": [{"address": ["3Fkx2TFdcHoab4xGgSjhAVh5YBPvbBWjNL"], "value": 340000, "txType": 5}], "timestamp": 1231639200, "totalInput": 345000, "totalOutput": 340000}
{"txIdGroup": 0, "txId": 10, "txHash": "ab403d5d27987d2a3dfc724e359870c6644b40e497bdc0589a033220fe15429d", "coinbase": false, "coinjoin": false, "blockId": 4, "inputs": [{"address": ["1Fufjpf9RM2aQsGedhSpbSCGRHrmLMJ7yY"], "value": 100000, "txType": 3}, {"address": ["1231PgW8KbpwKkvACPhp13fcL6fM5sKGvy"], "value": 1325000, "txType": 3}], "outputs": [{"address": ["1Fm1rJe1CtGuTKxWjQ4MyU7GEwrgkHYn5k"], "value": 1400000, "txType": 3}], "timestamp": 1231639200, "totalInput": 1425000, "totalOutput": 1400000}
{"txIdGroup": 0, "txId": 11, "txHash": "ab403d5d27987d2a3dfc724e359870c6644b40e497bdc0589a033220fe15429d", "coinbase": false, "coinjoin": false, "blockId": 4, "inputs": [{"address": ["ltc1qd42h5rlss8d54hpzvjpy408z2d8gpnj668wnt0"], "value": 100000, "txType": 8}, {"address": ["ltc1qq8gtfz5dvd838na8v86ehd3l98qjatpz47mayk"], "value": 1325000, "txType": 8}], "outputs": [{"address": ["ltc1qq8gtfz5dvd838na8v86ehd3l98qjatpz47mayk"], "value": 1400000, "txType": 3}], "timestamp": 1231639200, "totalInput": 1425000, "totalOutput": 1400000}
""",
    drop=[
        r".*Group",
    ],
)

exchange_rates = preprocess_inputs(
    """
{"date": "2009-01-03", "fiatValues": {"EUR": 150, "USD": 149.55}, "blockId": 1}
{"date": "2009-01-09", "fiatValues": {"EUR": 300, "USD": 302.67}, "blockId": 2}
{"date": "2009-01-10", "fiatValues": {"EUR": 600, "USD": 598.23}, "blockId": 3}
{"date": "2009-01-11", "fiatValues": {"EUR": 500, "USD": 498.18}, "blockId": 4}
"""
)


# src/test/resources/reference/address_txs.json
address_txs = preprocess_inputs(
    """
{"value": 2500000, "blockId": 1, "txId": 1, "addressIdGroup": 0, "addressId": 0, "isOutgoing": false}
{"value": -2500000, "blockId": 2, "txId": 3, "addressIdGroup": 0, "addressId": 0, "isOutgoing": true}
{"value": 2500000, "blockId": 2, "txId": 2, "addressIdGroup": 0, "addressId": 1, "isOutgoing": false}
{"value": -2500000, "blockId": 3, "txId": 5, "addressIdGroup": 0, "addressId": 1, "isOutgoing": true}
{"value": 1480000, "blockId": 2, "txId": 3, "addressIdGroup": 1, "addressId": 2, "isOutgoing": false}
{"value": -1480000, "blockId": 3, "txId": 5, "addressIdGroup": 1, "addressId": 2, "isOutgoing": true}
{"value": 495000, "blockId": 2, "txId": 3, "addressIdGroup": 1, "addressId": 3, "isOutgoing": false}
{"value": -150000, "blockId": 3, "txId": 6, "addressIdGroup": 1, "addressId": 3, "isOutgoing": true}
{"value": -345000, "blockId": 4, "txId": 9, "addressIdGroup": 1, "addressId": 3, "isOutgoing": true}
{"value": 495000, "blockId": 2, "txId": 3, "addressIdGroup": 2, "addressId": 4, "isOutgoing": false}
{"value": -495000, "blockId": 4, "txId": 8, "addressIdGroup": 2, "addressId": 4, "isOutgoing": true}
{"value": 2500000, "blockId": 3, "txId": 4, "addressIdGroup": 2, "addressId": 5, "isOutgoing": false}
{"value": -2500000, "blockId": 4, "txId": 8, "addressIdGroup": 2, "addressId": 5, "isOutgoing": true}
{"value": 1325000, "blockId": 3, "txId": 5, "addressIdGroup": 3, "addressId": 6, "isOutgoing": false}
{"value": -1325000, "blockId": 4, "txId": 10, "addressIdGroup": 3, "addressId": 6, "isOutgoing": true}
{"value": 100000, "blockId": 3, "txId": 5, "addressIdGroup": 3, "addressId": 7, "isOutgoing": false}
{"value": -100000, "blockId": 4, "txId": 10, "addressIdGroup": 3, "addressId": 7, "isOutgoing": true}
{"value": 2350000, "blockId": 3, "txId": 5, "addressIdGroup": 4, "addressId": 8, "isOutgoing": false}
{"value": 140000, "blockId": 3, "txId": 6, "addressIdGroup": 4, "addressId": 9, "isOutgoing": false}
{"value": 340000, "blockId": 4, "txId": 9, "addressIdGroup": 4, "addressId": 9, "isOutgoing": false}
{"value": 2500000, "blockId": 4, "txId": 7, "addressIdGroup": 5, "addressId": 10, "isOutgoing": false}
{"value": 4300000, "blockId": 4, "txId": 8, "addressIdGroup": 5, "addressId": 11, "isOutgoing": false}
{"value": 149000, "blockId": 4, "txId": 8, "addressIdGroup": 6, "addressId": 12, "isOutgoing": false}
{"value": 1400000, "blockId": 4, "txId": 10, "addressIdGroup": 6, "addressId": 13, "isOutgoing": false}
{"value": -100000, "blockId": 4, "txId": 11, "addressIdGroup": 6, "addressId": 14, "isOutgoing": true}
{"value": 75000, "blockId": 4, "txId": 11, "addressIdGroup": 6, "addressId": 15, "isOutgoing": false}
""",
    drop=[r".*Group", "blockId"],
)

# src/test/resources/reference/address_ids.json
address_ids = preprocess_inputs(
    """
{"addressId": 0, "address": "138cWsiAGpW9yqfjMVCCsFcnaiSHyoWMnJ"}
{"addressId": 1, "address": "1HxbaBMF2jXBVnagoHvaA6rLxmaYL8gb8T"}
{"addressId": 2, "address": "1FDCgJ8m2xDyVmYuankk13XReVC2Zvs5cz"}
{"addressId": 3, "address": "1FAkhqm95YnV5Mi7Q5j2Wb8CkbK7Z9zpyB"}
{"addressId": 4, "address": "3Kawbc5pkpQPfmaoGCBiaw5t2uK6WBoVVe"}
{"addressId": 5, "address": "12751KvnaTTPMJbHbbercdTV48kL81BjX8"}
{"addressId": 6, "address": "1231PgW8KbpwKkvACPhp13fcL6fM5sKGvy"}
{"addressId": 7, "address": "1Fufjpf9RM2aQsGedhSpbSCGRHrmLMJ7yY"}
{"addressId": 8, "address": "1iYSYHTpr2wMShaXTTNUzMohkpuV5p5ep"}
{"addressId": 9, "address": "3Fkx2TFdcHoab4xGgSjhAVh5YBPvbBWjNL"}
{"addressId": 10, "address": "1CjPR7Z5ZSyWk6WtXvSFgkptmpoi4UM9BC"}
{"addressId": 11, "address": "1AGMAXWELLayCyS1vkLXEszESHEcB3LWqa"}
{"addressId": 12, "address": "3JX79i9xSSmLEDZ8WrUFsy3WqrpvPoQmPv"}
{"addressId": 13, "address": "1Fm1rJe1CtGuTKxWjQ4MyU7GEwrgkHYn5k"}
{"addressId": 14, "address": "ltc1qd42h5rlss8d54hpzvjpy408z2d8gpnj668wnt0"}
{"addressId": 15, "address": "ltc1qq8gtfz5dvd838na8v86ehd3l98qjatpz47mayk"}
"""
)

# src/test/resources/reference/address_relations.json
address_relations = preprocess_inputs(
    """
{"srcAddressIdGroup": 0, "srcAddressId": 0, "dstAddressIdGroup": 1, "dstAddressId": 2, "noTransactions": 1, "estimatedValue": {"value": 1480000, "fiatValues": [4.44, 4.48]}, "txList": [3]}
{"srcAddressIdGroup": 0, "srcAddressId": 0, "dstAddressIdGroup": 1, "dstAddressId": 3, "noTransactions": 1, "estimatedValue": {"value": 495000, "fiatValues": [1.49, 1.5]}, "txList": [3]}
{"srcAddressIdGroup": 0, "srcAddressId": 0, "dstAddressIdGroup": 2, "dstAddressId": 4, "noTransactions": 1, "estimatedValue": {"value": 495000, "fiatValues": [1.49, 1.5]}, "txList": [3]}
{"srcAddressIdGroup": 0, "srcAddressId": 1, "dstAddressIdGroup": 3, "dstAddressId": 6, "noTransactions": 1, "estimatedValue": {"value": 832286, "fiatValues": [4.99, 4.98]}, "txList": [5]}
{"srcAddressIdGroup": 1, "srcAddressId": 2, "dstAddressIdGroup": 3, "dstAddressId": 6, "noTransactions": 1, "estimatedValue": {"value": 492714, "fiatValues": [2.96, 2.95]}, "txList": [5]}
{"srcAddressIdGroup": 0, "srcAddressId": 1, "dstAddressIdGroup": 3, "dstAddressId": 7, "noTransactions": 1, "estimatedValue": {"value": 62814, "fiatValues": [0.38, 0.38]}, "txList": [5]}
{"srcAddressIdGroup": 1, "srcAddressId": 2, "dstAddressIdGroup": 3, "dstAddressId": 7, "noTransactions": 1, "estimatedValue": {"value": 37186, "fiatValues": [0.22, 0.22]}, "txList": [5]}
{"srcAddressIdGroup": 0, "srcAddressId": 1, "dstAddressIdGroup": 4, "dstAddressId": 8, "noTransactions": 1, "estimatedValue": {"value": 1476131, "fiatValues": [8.86, 8.83]}, "txList": [5]}
{"srcAddressIdGroup": 1, "srcAddressId": 2, "dstAddressIdGroup": 4, "dstAddressId": 8, "noTransactions": 1, "estimatedValue": {"value": 873869, "fiatValues": [5.24, 5.23]}, "txList": [5]}
{"srcAddressIdGroup": 1, "srcAddressId": 3, "dstAddressIdGroup": 4, "dstAddressId": 9, "noTransactions": 2, "estimatedValue": {"value": 480000, "fiatValues": [2.54, 2.53]}, "txList": [9, 6]}
{"srcAddressIdGroup": 2, "srcAddressId": 4, "dstAddressIdGroup": 5, "dstAddressId": 11, "noTransactions": 1, "estimatedValue": {"value": 473526, "fiatValues": [2.37, 2.36]}, "txList": [8]}
{"srcAddressIdGroup": 2, "srcAddressId": 5, "dstAddressIdGroup": 5, "dstAddressId": 11, "noTransactions": 1, "estimatedValue": {"value": 2391546, "fiatValues": [11.96, 11.91]}, "txList": [8]}
{"srcAddressIdGroup": 2, "srcAddressId": 4, "dstAddressIdGroup": 6, "dstAddressId": 12, "noTransactions": 1, "estimatedValue": {"value": 16408, "fiatValues": [0.08, 0.08]}, "txList": [8]}
{"srcAddressIdGroup": 2, "srcAddressId": 5, "dstAddressIdGroup": 6, "dstAddressId": 12, "noTransactions": 1, "estimatedValue": {"value": 82870, "fiatValues": [0.41, 0.41]}, "txList": [8]}
{"srcAddressIdGroup": 3, "srcAddressId": 6, "dstAddressIdGroup": 6, "dstAddressId": 13, "noTransactions": 1, "estimatedValue": {"value": 1301754, "fiatValues": [6.51, 6.49]}, "txList": [10]}
{"srcAddressIdGroup": 3, "srcAddressId": 7, "dstAddressIdGroup": 6, "dstAddressId": 13, "noTransactions": 1, "estimatedValue": {"value": 98246, "fiatValues": [0.49, 0.49]}, "txList": [10]}
{"srcAddressIdGroup": 3, "srcAddressId": 14, "dstAddressIdGroup": 6, "dstAddressId": 15, "noTransactions": 1, "estimatedValue": {"value": 75000, "fiatValues": [0.38, 0.37]}, "txList": [11]}
""",
    drop=[r".*Group"],
)


def resolve_address_id(item, addr_id_lookup):
    prefix = "address_id"
    candidates = [x for x in item.keys() if x.endswith("address_id")]
    for c in candidates:
        item[c.replace(prefix, "address")] = addr_id_lookup[item.pop(c)]
    return item.copy()


def get_address_id_lookup():
    return {x["address_id"]: x["address"] for x in address_ids}


def rename_address_to_identifier(d):
    d["identifier"] = d.pop("address")
    return d


def get_txs():
    return [dict_to_dataobject(tx) for tx in txs]


def get_atxs():
    id_to_adr = get_address_id_lookup()
    return [
        dict_to_dataobject(
            rename_address_to_identifier(resolve_address_id(adr_tx, id_to_adr))
        )
        for adr_tx in address_txs
    ]


def get_exchange_rates_per_block():
    return {x["block_id"]: x["fiat_values"].values() for x in exchange_rates}


def get_arel():
    id_to_adr = get_address_id_lookup()
    return [
        dict_to_dataobject(resolve_address_id(adr_rel, id_to_adr))
        for adr_rel in address_relations
    ]



import os

from graphsenselib.utils import (
    batch,
    batch_date,
    bytes_to_hex,
    first_or_default,
    generate_date_range_days,
    remove_prefix,
    strip_0x,
    subkey_exists,
    subkey_get,
    to_int,
)
from graphsenselib.utils.errorhandling import CrashRecoverer


def test_btoh_works(capsys):
    assert bytes_to_hex(b"") is None
    assert bytes_to_hex(b"asdfasdf") == "6173646661736466"


def test_strip_0x_works1():
    assert strip_0x("0xa9059cbb") == "a9059cbb"


def test_strip_0x_works2():
    assert strip_0x("a9059cbb") == "a9059cbb"
    assert strip_0x(None) is None


def test_to_int1():
    assert to_int(0) == 0
    assert to_int("1111") == 1111
    assert to_int("0xf") == 15
    assert to_int("0x10") == 16


def test_crash_recoverer():
    file = "/tmp/test_graphsense_lib_crashrecoverer.err"
    if os.path.exists(file):
        os.remove(file)
    cr = CrashRecoverer(file)

    try:
        with cr.enter_critical_section({"int": 1}):
            raise NotImplementedError("test")
    except NotImplementedError:
        assert cr.is_in_recovery_mode()
        assert cr.get_recovery_hint() == {
            "int": 1,
            "exception": "test",
            "exception_type": "NotImplementedError",
        }

    cr = CrashRecoverer(file)
    assert cr.is_in_recovery_mode()
    assert cr.get_recovery_hint() == {
        "int": 1,
        "exception": "test",
        "exception_type": "NotImplementedError",
    }

    try:
        with cr.enter_critical_section({}):
            pass
    except ValueError:
        pass

    assert cr.get_recovery_hint() == {
        "int": 1,
        "exception": "test",
        "exception_type": "NotImplementedError",
    }

    cr.leave_recovery_mode()
    assert not cr.is_in_recovery_mode()
    assert not os.path.exists(file)


def test_batch_works():
    assert [list(b) for b in batch(range(763638, 763639 + 1), n=1)] == [
        [763638],
        [763639],
    ]
    assert [list(b) for b in batch(range(763638, 763639 + 1), n=2)] == [
        [763638, 763639]
    ]
    assert [list(b) for b in batch(range(763638, 763640 + 1), n=2)] == [
        [763638, 763639],
        [763640],
    ]


def test_subkey_exists():
    assert subkey_exists({}, ["abc"]) is False
    assert subkey_exists({"abc": None}, ["abc"]) is True
    assert subkey_exists({"abc": []}, ["abc", "cbd"]) is False
    assert subkey_exists({"abc": {"cbd": 1}}, ["abc", "cbd"]) is True
    assert subkey_exists({"abc": {"cbd": None}}, ["abc", "cbd"]) is True


def test_subkey_get():
    assert subkey_get({}, ["abc"]) is None
    assert subkey_get({"abc": None}, ["abc"]) is None
    assert subkey_get({"abc": []}, ["abc", "cbd"]) is None
    assert subkey_get({"abc": {"cbd": 1}}, ["abc", "cbd"]) == 1
    assert subkey_get({"abc": {"cbd": "string"}}, ["abc", "cbd"]) == "string"
    assert (
        subkey_get({"abc": {"cbd": ["a", "b"], "bbb": [1, 2, 3]}}, ["abc", "cbd", "1"])
        == "b"
    )
    assert subkey_get(
        {"abc": {"cbd": ["a", "b"], "bbb": [1, 2, 3]}}, ["abc", "bbb"]
    ) == [1, 2, 3]


def test_first_or_default():
    assert first_or_default([1, 2, 3], lambda x: x > 2, default=10) == 3
    assert first_or_default([1, 2, 3], lambda x: x > 5, default=10) == 10
    assert first_or_default([1, 2, 3], lambda x: x > 5, default=None) is None


def test_remove_prefix():
    assert remove_prefix("0xa9059cbb", "0x") == "a9059cbb"
    assert remove_prefix("0xa9059cbb", "a0x") == "0xa9059cbb"


def test_date_range_works():
    from datetime import date

    a = date.fromisoformat("2010-03-12")
    b = date.fromisoformat("2024-05-19")

    batches = list(batch_date(a, b, days=180))

    deltas = [y - x for x, y in batches]

    assert all(d.days == 180 for d in deltas[:-1])

    assert batches[0][0] == a
    assert batches[-1][1] == b
    assert (sum(d.days for d in deltas) + len(deltas) - 1) == (b - a).days

    gen_dates = list(generate_date_range_days(a, b))

    assert gen_dates[0] == a
    assert gen_dates[-1] == b

    assert len(gen_dates) == (b - a).days + 1



from __future__ import annotations

from dataclasses import dataclass
from functools import reduce
from typing import List

from cassandra.cqlengine.usertype import UserType

from graphsenselib.deltaupdate.update.generic import (
    DeltaScalar,
    DeltaUpdate,
    DeltaValue,
    merge_asset_dicts,
    minusone_respecting_function,
)
from graphsenselib.utils import group_by, groupby_property


@dataclass
class EntityDeltaAccount(DeltaUpdate):
    identifier: str
    total_received: DeltaValue
    total_spent: DeltaValue
    total_tokens_received: dict[str, DeltaValue]
    total_tokens_spent: dict[str, DeltaValue]
    first_tx_id: int
    last_tx_id: int
    no_incoming_txs: int
    no_outgoing_txs: int
    no_incoming_txs_zero_value: int
    no_outgoing_txs_zero_value: int

    @classmethod
    def from_db(Cls, db_row):
        identifier = db_row.address

        # empty mapping is None in cassandra
        # python cassandra driver saves {} as None, so we dont
        # need to worry about empty dicts not
        # fitting into cassandra later
        if db_row.total_tokens_spent is None:
            total_tokens_spent = {}
        else:
            total_tokens_spent = {
                k: DeltaValue.from_db(v) for k, v in db_row.total_tokens_spent.items()
            }

        if db_row.total_tokens_received is None:
            total_tokens_received = {}
        else:
            total_tokens_received = {
                k: DeltaValue.from_db(v)
                for k, v in db_row.total_tokens_received.items()
            }

        return Cls(
            identifier=identifier,
            total_received=DeltaValue.from_db(db_row.total_received),
            total_spent=DeltaValue.from_db(db_row.total_spent),
            total_tokens_received=total_tokens_received,
            total_tokens_spent=total_tokens_spent,
            first_tx_id=db_row.first_tx_id,
            last_tx_id=db_row.last_tx_id,
            no_incoming_txs=db_row.no_incoming_txs,
            no_outgoing_txs=db_row.no_outgoing_txs,
            no_incoming_txs_zero_value=db_row.no_incoming_txs_zero_value,
            no_outgoing_txs_zero_value=db_row.no_outgoing_txs_zero_value,
        )

    def merge(self, other_delta):
        assert self.identifier == other_delta.identifier

        # self and other total_tokens_received
        # may not have the same keys, fix the following:
        total_tokens_received = merge_asset_dicts(
            self.total_tokens_received, other_delta.total_tokens_received
        )

        total_tokens_spent = merge_asset_dicts(
            self.total_tokens_spent, other_delta.total_tokens_spent
        )

        return EntityDeltaAccount(
            identifier=self.identifier,
            total_received=self.total_received.merge(other_delta.total_received),
            total_spent=self.total_spent.merge(other_delta.total_spent),
            total_tokens_received=total_tokens_received,
            total_tokens_spent=total_tokens_spent,
            first_tx_id=minusone_respecting_function(
                self.first_tx_id, other_delta.first_tx_id, min
            ),
            last_tx_id=minusone_respecting_function(
                self.last_tx_id, other_delta.last_tx_id, max
            ),
            no_incoming_txs=self.no_incoming_txs + other_delta.no_incoming_txs,
            no_outgoing_txs=self.no_outgoing_txs + other_delta.no_outgoing_txs,
            no_incoming_txs_zero_value=self.no_incoming_txs_zero_value
            + other_delta.no_incoming_txs_zero_value,
            no_outgoing_txs_zero_value=self.no_outgoing_txs_zero_value
            + other_delta.no_outgoing_txs_zero_value,
        )


@dataclass
class RawEntityTxAccount:
    identifier: str
    is_outgoing: bool
    tx_id: int
    tx_reference: UserType
    block_id: int
    value: int
    token_values: dict[str, int]


@dataclass
class RelationDeltaAccount(DeltaUpdate):
    src_identifier: bytes
    dst_identifier: bytes
    no_transactions: int
    value: DeltaValue
    token_values: dict[str, DeltaValue]
    type: str  # noqa

    @classmethod
    def from_db(Cls, db_row):
        return Cls(
            src_identifier=getattr(db_row, "src_address"),
            dst_identifier=getattr(db_row, "dst_address"),
            no_transactions=db_row.no_transactions,
            value=DeltaValue.from_db(db_row.value),
            token_values={
                k: DeltaValue.from_db(v) for k, v in db_row.token_values.items()
            },
            type="from_db",
        )

    def merge(self, other_delta):
        assert self.src_identifier == other_delta.src_identifier
        assert self.dst_identifier == other_delta.dst_identifier

        token_values = merge_asset_dicts(self.token_values, other_delta.token_values)

        return RelationDeltaAccount(
            src_identifier=self.src_identifier,
            dst_identifier=self.dst_identifier,
            value=self.value.merge(other_delta.value),
            token_values=token_values,
            no_transactions=self.no_transactions + other_delta.no_transactions,
            type="merged",
        )


@dataclass
class BalanceDelta(DeltaUpdate):
    identifier: int
    asset_balances: dict[str, DeltaScalar]

    @classmethod
    def from_db(Cls, identifier, db_row_list):
        if len(db_row_list) == 0:
            return Cls(
                identifier=identifier,
                asset_balances={},
            )

        asset_balances = {x.currency: DeltaScalar(x.balance) for x in db_row_list}
        return Cls(
            identifier=identifier,
            asset_balances=asset_balances,
        )

    def merge(self, other_delta):
        assert self.identifier == other_delta.identifier

        asset_balances = {
            k: self.asset_balances.get(k, DeltaScalar(0)).merge(
                other_delta.asset_balances.get(k, DeltaScalar(0))
            )
            for k in set(self.asset_balances.keys())
            | set(other_delta.asset_balances.keys())
        }
        return BalanceDelta(
            identifier=self.identifier,
            asset_balances=asset_balances,
        )

    def left_join(self, other_delta):
        assert self.identifier == other_delta.identifier
        asset_balances = {
            k: self.asset_balances.get(k, DeltaScalar(0)).merge(
                other_delta.asset_balances.get(k, DeltaScalar(0))
            )
            for k in self.asset_balances.keys()
        }
        return BalanceDelta(
            identifier=self.identifier,
            asset_balances=asset_balances,
        )


@dataclass
class DbDeltaAccount:
    entity_updates: List[EntityDeltaAccount]
    new_entity_txs: List[RawEntityTxAccount]
    relation_updates: List[RelationDeltaAccount]
    balance_updates: List[BalanceDelta]

    def concat(self, other):
        return DbDeltaAccount(
            entity_updates=self.entity_updates + other.entity_updates,
            new_entity_txs=self.new_entity_txs + other.new_entity_txs,
            relation_updates=self.relation_updates + other.relation_updates,
            balance_updates=self.balance_updates + other.balance_updates,
        )

    @staticmethod
    def merge(change_sets: List[DbDeltaAccount]) -> "DbDeltaAccount":
        return reduce(lambda x, y: x.concat(y), change_sets).compress()

    def compress(self):
        grouped = groupby_property(
            self.entity_updates, "identifier", sort_by="first_tx_id"
        )
        entity_updates_merged = {
            k: reduce(lambda x, y: x.merge(y), v) for k, v in grouped.items()
        }
        assert len(entity_updates_merged.keys()) == len(
            set(entity_updates_merged.keys())
        )

        grouped = group_by(
            self.relation_updates, lambda x: (x.src_identifier, x.dst_identifier)
        )
        relations_updates_merged = {
            (src, dst): reduce(lambda x, y: x.merge(y), v)
            for (src, dst), v in grouped.items()
        }

        grouped = group_by(self.balance_updates, lambda x: x.identifier)
        balance_updates_merged = {
            k: reduce(lambda x, y: x.merge(y), v) for k, v in grouped.items()
        }

        return DbDeltaAccount(
            entity_updates=sorted(
                entity_updates_merged.values(),
                key=lambda x: (x.first_tx_id, x.last_tx_id),
            ),
            new_entity_txs=self.new_entity_txs,
            relation_updates=list(relations_updates_merged.values()),
            balance_updates=list(balance_updates_merged.values()),
        )



from dataclasses import dataclass

import pandas as pd
from eth_abi import decode_single
from eth_utils import function_abi_to_4byte_selector, to_hex
from web3 import Web3


@dataclass
class TokenTransfer:
    from_address: bytes
    to_address: bytes
    value: int
    asset: str
    decimals: int
    coin_equivalent: int
    usd_equivalent: float
    block_id: int
    tx_hash: bytes
    log_index: int


class ERC20Decoder:
    def __init__(self, currency: str, supported_tokens: pd.DataFrame):
        self.w3 = Web3()

        self.token_transfer_event_abi = {
            "anonymous": False,
            "inputs": [
                {"indexed": True, "name": "from", "type": "address"},
                {"indexed": True, "name": "to", "type": "address"},
                {"indexed": False, "name": "value", "type": "uint256"},
            ],
            "name": "Transfer",
            "type": "event",
        }

        self.token_transfer_event_selector = self.get_event_selector(
            self.token_transfer_event_abi
        )

        self.currency = currency.upper()
        self.supported_tokens = supported_tokens

    def get_event_selector(self, event_abi):
        return to_hex(function_abi_to_4byte_selector(event_abi))

    def log_to_transfer(self, log):
        if "0x" + log.address.hex() in self.supported_tokens["token_address"].values:
            return self.decode_transfer(log)

    def decode_transfer(self, log):
        if "0x" + log.topics[0].hex()[:8] == self.token_transfer_event_selector:
            if (
                "0x" + log.address.hex()
                not in self.supported_tokens["token_address"].values
            ):
                raise Exception(
                    "Unsupported token, use the log_to_transfer function instead"
                )

            try:
                sender = bytes.fromhex(
                    self.w3.toChecksumAddress(decode_single("address", log.topics[1]))[
                        2:
                    ]
                )
                recipient = bytes.fromhex(
                    self.w3.toChecksumAddress(decode_single("address", log.topics[2]))[
                        2:
                    ]
                )
                value = decode_single("uint256", log.data)
                mask = (
                    self.supported_tokens["token_address"] == "0x" + log.address.hex()
                )
                asset = self.supported_tokens[mask]["currency_ticker"].values[0]
                peg = self.supported_tokens[mask]["peg_currency"].values[0]
                coin_equivalent = peg == self.currency
                usd_equivalent = peg == "USD"
                decimals = self.supported_tokens[mask]["decimals"].values[0]

                return TokenTransfer(
                    from_address=sender,
                    to_address=recipient,
                    value=value,
                    asset=asset,
                    decimals=decimals,
                    coin_equivalent=coin_equivalent,
                    usd_equivalent=usd_equivalent,
                    block_id=log.block_id,
                    tx_hash=log.tx_hash,
                    log_index=log.log_index,
                )
            except Exception:
                return None  # cant be decoded
        else:
            return None  # not a transfer event



import json
import os
from contextlib import contextmanager


@contextmanager
def cr_critical_section(recovery_hint: dict, crash_recoverer):
    try:
        yield
    except Exception as e:
        recovery_hint["exception"] = str(e)
        recovery_hint["exception_type"] = type(e).__name__
        crash_recoverer.enter_recovery_mode(recovery_hint)
        raise e


def get_exception_digest(ex):
    return f"{type(ex).__name__}: {str(ex)}"


class CrashRecoverer:
    """
    Class permanently stores (in a file) a crash hint if a program crashes in a critical
    If the class is reinstantiated with the same identifier and a crash hint is present
    the class does not allow the program to enter the critical section again until
    the crash and the underlying inconsistency is resolved
    (e.g. by calling leave_recover_mode) or by deleting the crash hint file.

    Crash hints are user specified and in the optimal case should contain enough
    information for the program to recover from crashes automatically on the
    next run.
    """

    def __init__(self, crashfile_name: str):
        self._crashfile = os.path.expanduser(crashfile_name)

    def is_in_recovery_mode(self) -> bool:
        return os.path.exists(self._crashfile)

    def enter_critical_section(self, recovery_hint: dict):
        """
        test if recovery hint is json serializable
        to avoid that serialisation fails in case of an error
        """
        json.dumps(recovery_hint)
        if self.is_in_recovery_mode():
            raise ValueError("We are already in recovery mode.")
        return cr_critical_section(recovery_hint, self)

    def get_recovery_hint(self) -> dict:
        with open(self._crashfile, "r") as f:
            return json.load(f)

    def get_recovery_hint_filename(self) -> str:
        return self._crashfile

    def enter_recovery_mode(self, recovery_hint: dict):
        if self.is_in_recovery_mode():
            raise ValueError("We are already in recovery mode.")
        with open(os.path.expanduser(self._crashfile), "w") as f:
            json.dump(recovery_hint, f)

    def leave_recovery_mode(self):
        if self.is_in_recovery_mode():
            os.remove(self._crashfile)



from typing import Dict, Tuple, Union

from . import group_by

DEFAULT_KEY_ENCODERS = {
    bytes: lambda x: x.hex(),
    int: lambda x: str(x),
    str: lambda x: x,
}


class TableBasedCache:
    def __init__(
        self,
        internal_key_value_cache: Dict,
        table_delimiter: str = "|",
        key_encoder=DEFAULT_KEY_ENCODERS,
    ):
        self.internal_key_value_cache = internal_key_value_cache
        self.delim = table_delimiter
        self.key_encoder = key_encoder

    def get_key(self, table: str, key: Union[bytes, int, str]):
        assert self.delim not in table
        kt = type(key)
        if kt not in self.key_encoder:
            raise ValueError(f"Don't know how to encode key of type {kt}")
        ekey = self.key_encoder[kt](key)
        return f"{ekey}{self.delim}{table}"

    def __delitem__(self, kv: Tuple[str, Union[bytes, int, str]]):
        table, key = kv
        self.delete_item(table, key)

    def delete_item(self, table: str, key: Union[bytes, int, str]):
        self.internal_key_value_cache.delete(self.get_key(table, key), retry=True)

    def put_item(self, table: str, key: Union[bytes, int, str], item):
        self.internal_key_value_cache[self.get_key(table, key)] = item

    def put_items(self, table: str, items: Tuple[str, object]):
        for k, v in items:
            self.put_item(table, k, v)

    def put_items_keyed_by(self, table: str, items: Tuple[str, object], key: str):
        by_key = group_by(items, key=lambda x: x[key])

        self.put_items(table, by_key.items())

    def __setitem__(self, kv: Tuple[str, Union[bytes, int, str]], data):
        table, key = kv
        self.put_item(table, key, data)

    def __getitem__(self, kv: Tuple[str, Union[bytes, int, str]]):
        table, key = kv
        return self.get_item(table, key)

    def get(self, kv: Tuple[str, Union[bytes, int, str]], default=None):
        try:
            return self[kv]
        except KeyError:
            return default

    def get_item(self, table: str, key: str):
        return self.internal_key_value_cache[self.get_key(table, key)]



import sys


def info(extype, value, tb):
    if hasattr(sys, "ps1") or not sys.stderr.isatty():
        # we are in interactive mode or we don't have a tty-like
        # device, so we call the default hook
        sys.__excepthook__(extype, value, tb)
    else:
        import pdb
        import traceback

        # we are NOT in interactive mode, print the exception...
        traceback.print_exception(extype, value, tb)
        print
        # ...then start the debugger in post-mortem mode.
        pdb.post_mortem(tb)


sys.excepthook = info



from typing import Optional, Union

from .generic import remove_prefix

# from typeguard import typechecked


# @typechecked
def hex_str_to_bytes(hex_str: str) -> bytes:
    return bytes.fromhex(hex_str)


# @typechecked
def bytes_to_hex(b: bytes) -> Optional[str]:
    r = bytes(b).hex()
    return r if len(r) > 0 else None


# @typechecked
def is_hex_string(string: Optional[str]) -> bool:
    return string is not None and string.startswith("0x") and len(string) >= 2


# @typechecked
def strip_0x(string: Optional[str]) -> Optional[str]:
    return remove_prefix(string, "0x") if is_hex_string(string) else string


# @typechecked
def to_int(string: Union[str, int]) -> int:
    if type(string) == int:
        return string

    if is_hex_string(string):
        return int(string, 16)
    else:
        return int(string)


# @typechecked
def hex_to_bytes(hex_str: Optional[str]) -> Optional[bytes]:
    """Convert hexstring (starting with 0x) to bytearray."""
    return bytes.fromhex(strip_0x(hex_str)) if hex_str is not None else None



from graphsenselib.config import config, get_reorg_backoff_blocks


def test_config_is_not_loaded_by_default():
    # But real config should not be automatically loaded.
    assert config.is_loaded() is False


def test_get_approx_reorg_backoff_blocks():
    assert get_reorg_backoff_blocks("eth") == 70



import pytest

from graphsenselib.utils.accountmodel import hex_str_to_bytes
from graphsenselib.utils.tron import (
    add_tron_prefix,
    decode_note,
    evm_to_bytes,
    evm_to_tron_address_string,
    partial_tron_to_partial_evm,
    tron_address_equal,
    tron_address_to_bytes,
    tron_address_to_evm_string,
    tron_address_to_legacy,
    tron_address_to_legacy_string,
)

# first is evm address and second is tron address
correct_encodings = [
    (
        "414450cf8c8b6a8229b7f628e36b3a658e84441b6f",
        "TGCRkw1Vq759FBCrwxkZGgqZbRX1WkBHSu",
    ),
    ("4450cf8c8b6a8229b7f628e36b3a658e84441b6f", "TGCRkw1Vq759FBCrwxkZGgqZbRX1WkBHSu"),
    (
        "0x4450cf8c8b6a8229b7f628e36b3a658e84441b6f",
        "TGCRkw1Vq759FBCrwxkZGgqZbRX1WkBHSu",
    ),
    (
        "0xe552f6487585c2b58bc2c9bb4492bc1f17132cd0",
        "TWsm8HtU2A5eEzoT8ev8yaoFjHsXLLrckb",
    ),
    (
        "0x5095d4f4d26ebc672ca12fc0e3a48d6ce3b169d2",
        "THKJYuUmMKKARNf7s2VT51g5uPY6KEqnat",
    ),
]


def test_add_tron_prefix():
    # prefix must not be specified
    assert hex_str_to_bytes(
        "414450cf8c8b6a8229b7f628e36b3a658e84441b6f"
    ) == add_tron_prefix(
        hex_str_to_bytes("414450cf8c8b6a8229b7f628e36b3a658e84441b6f"[2:])
    )


def test_evm_to_tron_address():
    for evm, tron in correct_encodings:
        assert evm_to_tron_address_string(evm) == tron


def test_tron_address_equality():
    for evm, tron in correct_encodings:
        assert evm_to_bytes(evm) == tron_address_to_bytes(tron)
        assert tron_address_equal(evm, tron)


def test_checksum_error():
    with pytest.raises(ValueError):
        # introduced an error in the checksum postfix qnat -> qnaf
        tron_address_to_legacy("THKJYuUmMKKARNf7s2VT51g5uPY6KEqnaf")

    assert (
        tron_address_to_legacy_string(
            "THKJYuUmMKKARNf7s2VT51g5uPY6KEqnaf", validate=False
        )
        == "0x5095d4f4d26ebc672ca12fc0e3a48d6ce3b169d2"
    )

    assert tron_address_to_legacy(
        "THKJYuUmMKKARNf7s2VT51g5uPY6KEqnaf", validate=False
    ) == evm_to_bytes("0x5095d4f4d26ebc672ca12fc0e3a48d6ce3b169d2")


def test_tron_bijection():
    original_evm = "0x5095d4f4d26ebc672ca12fc0e3a48d6ce3b169d2"
    original_tron = "THKJYuUmMKKARNf7s2VT51g5uPY6KEqnat"

    f = evm_to_tron_address_string
    g = tron_address_to_evm_string

    assert g(f(original_evm)) == original_evm
    assert f(g(original_tron)) == original_tron


def test_decode_note():
    assert decode_note("63616c6c") == "call"


def test_partial_prefix_conversion():
    eth_style = "0x5095d4f4d26ebc672ca12fc0e3a48d6ce3b169d2"
    tron_style = "THKJYuUmMKKARNf7s2VT51g5uPY6KEqnat"

    for slice_len in range(34):
        actual = eth_style[:slice_len]
        converted = partial_tron_to_partial_evm(tron_style[:slice_len])
        assert actual == converted

    for slice_len in range(34, 43):
        actual = eth_style[:slice_len]
        converted = partial_tron_to_partial_evm(tron_style[:slice_len])
        assert actual == converted[:slice_len]



import logging
from datetime import datetime
from typing import Optional

import click
from eth_hash.auto import keccak

from ..cli.common import require_currency, require_environment
from ..config import currency_to_schema_type, supported_base_currencies
from ..datatypes.abi import decode_db_logs, decoded_log_to_str
from ..utils.accountmodel import hex_str_to_bytes, hex_to_bytes, is_hex_string, strip_0x
from ..utils.console import console
from .factory import DbFactory
from .trace import trace as trace_it

logger = logging.getLogger(__name__)


@click.group()
def db_cli():
    pass


@db_cli.group()
def db():
    """Query related functions."""
    pass


@db_cli.group()
def trace():
    """trace related functions."""
    pass


@db.command("state")
@require_environment()
@require_currency(required=False)
def state(env, currency):
    """Prints the current state of the graphsense database.
    \f
    Args:
        env (str): Environment to work on
        currency (str): currency to work on
    """
    currencies = supported_base_currencies if currency is None else [currency]
    for cur in currencies:
        console.rule(f"{cur}")
        with DbFactory().from_config(env, cur) as db:
            hb_ft = db.transformed.get_highest_block_fulltransform()
            hb_raw = db.raw.get_highest_block()
            start_block = (db.transformed.get_highest_block_delta_updater() or -1) + 1
            transformed_exists = db.transformed.exists()
            if transformed_exists:
                latest_address_id = db.transformed.get_highest_address_id()
                latest_clstr_id = db.transformed.get_highest_cluster_id()
            else:
                latest_address_id = 0
                latest_clstr_id = 0
            console.print(f"Last addr id:       {latest_address_id:12}")
            if latest_clstr_id is not None:
                console.print(f"Last cltr id:       {latest_clstr_id:12}")
            console.print(f"Raw     Config:      {db.raw.get_configuration()}")
            if transformed_exists:
                console.print(
                    f"Transf. Config:      {db.transformed.get_configuration()}"
                )
            end_block = db.raw.find_highest_block_with_exchange_rates()
            console.print(
                f"Last delta-transform: {(start_block -1):10}"
                f" ({db.raw.get_block_timestamp(start_block - 1)})"
            )
            console.print(
                f"Last raw block:       {hb_raw:10}"
                f" ({db.raw.get_block_timestamp(hb_raw)})"
            )
            console.print(
                f"Last raw block:       {end_block:10}"
                f" ({db.raw.get_block_timestamp(end_block)}) "
                "(with exchanges rates)."
            )
            if transformed_exists:
                console.print(
                    f"Transf. behind raw:   {(end_block - (start_block - 1)):10}"
                    f" ({db.raw.get_block_timestamp(end_block - (start_block - 1))})"
                    " (delta-transform)"
                )
                console.print(
                    f"Transf. behind raw:   {(end_block - hb_ft):10}"
                    f" ({db.raw.get_block_timestamp((end_block - hb_ft))})"
                    " (full-transform)"
                )


@db.group()
def block():
    """Special db query functions regarding blocks."""
    pass


@block.command("get-ts")
@require_environment()
@require_currency(required=False)
@click.option(
    "-b",
    "--block",
    type=int,
    required=True,
    help="block to get the ts for",
)
def get_ts(env: str, currency: str, block: int):
    """Summary
    Prints timestamps for the given block nr
    \f

    Args:
        env (str): Environment to work on
        currency (str): currency to work on
        block_id (int): Block to query
    """
    currencies = supported_base_currencies if currency is None else [currency]
    multi = len(currencies) > 1
    for cur in currencies:
        with DbFactory().from_config(env, cur) as db:
            timestamp = db.raw.get_block_timestamp(block)
            if multi:
                console.print(f"{cur}={timestamp}")
            else:
                console.print(f"{timestamp}")


@block.command("get-nr")
@require_environment()
@require_currency(required=False)
@click.option(
    "--date",
    type=click.DateTime(formats=["%Y-%m-%d %H:%M:%S"]),
    required=True,
    help="Date to get the block nr for.",
)
def get_nr(env: str, currency: str, date: datetime):
    """Summary
    Prints last blocknumber to include the given date.
    \f

    Args:
        env (str): Environment to work on
        currency (str): currency to work on
        block_id (int): Block to query
    """
    currencies = supported_base_currencies if currency is None else [currency]
    multi = len(currencies) > 1
    for cur in currencies:
        with DbFactory().from_config(env, cur) as db:
            nr = db.raw.find_block_nr_for_date(date)
            if nr is not None:
                tsp1 = db.raw.get_block_timestamp(nr + 1)
                ts = db.raw.get_block_timestamp(nr)
                tsm1 = db.raw.get_block_timestamp(nr - 1)
                logger.info(f"{tsm1} (-1) - {ts} ({nr}) - {tsp1} (+1)")
            if multi:
                console.print(f"{cur}={nr}")
            else:
                console.print(f"{nr}")


@db.group()
def logs():
    """Special db query functions regarding logs."""
    pass


@logs.command("get-decodeable-logs")
@require_environment()
@require_currency()
@click.option(
    "--start-block",
    type=int,
    required=True,
    help="Block to start fetching the decodable logs.",
)
@click.option(
    "--end-block",
    type=int,
    required=True,
    help="Block to stop fetching the decodable logs.",
)
@click.option(
    "--topic0",
    type=str,
    required=None,
    help="Log topic to fetch.",
)
@click.option(
    "--contract",
    type=str,
    required=None,
    help="Filter for Contract that produced the log.",
)
def get_logs(
    env: str,
    currency: str,
    start_block: int,
    end_block: int,
    topic0: Optional[str],
    contract: Optional[str],
):
    """Print all decodable logs for a given block
    Args:
        env (str): evironment
        currency (str): currency
        block (int): block
    """
    stype = currency_to_schema_type.get(currency, None)
    if stype == "account" or stype == "account_trx":
        with DbFactory().from_config(env, currency) as db:
            if topic0 is not None and is_hex_string(topic0):
                topic0 = hex_str_to_bytes(topic0)
            elif topic0 is not None:
                topic0 = keccak(topic0.encode("utf-8"))

            if contract is not None:
                contract = hex_str_to_bytes(strip_0x(contract))
            for b in range(start_block, end_block):
                for dlog, log in decode_db_logs(
                    db.raw.get_logs_in_block(b, topic0=topic0, contract=contract)
                ):
                    dlog_str = decoded_log_to_str(dlog)
                    print(f"{b}|{log.log_index}|0x{log.tx_hash.hex()}|{dlog_str}")
    else:
        print(
            f"Unsupported schema type {stype} for "
            f"currency {currency}. Only account is supported."
        )


@trace.command("events")
@require_environment()
@require_currency()
@click.option(
    "--q",
    type=str,
    required=True,
    help="Transaction hash or address",
)
@click.option(
    "--contract",
    type=str,
    required=False,
    help="Address of the contract to look at",
)
@click.option(
    "--start",
    type=str,
    required=False,
    help="Block or Date",
)
@click.option(
    "--end",
    type=str,
    required=False,
    help="Block or Date",
)
@click.option(
    "--names-file",
    type=str,
    default="./names.csv",
    help="File containing names for addresses and hashes etc.",
)
@click.option(
    "--output-format",
    type=str,
    default="csv",
    help="How to print results (csv, table etc.)",
)
def trace_transaction(
    env: str,
    currency: str,
    q: str,
    contract: str,
    start: str,
    end: str,
    names_file: str,
    output_format: str,
):
    """Print logs and filter
    Args:
        env (str): evironment
        currency (str): currency
        tx (str): tx
    """
    stype = currency_to_schema_type.get(currency, None)
    if stype == "account":
        with DbFactory().from_config(env, currency) as db:
            import os.path

            if os.path.isfile(names_file):
                with open(names_file) as file:
                    kvl = [line.split(":") for line in file.readlines()]
                    names = {a.strip(): b.strip() for a, b, *rest in kvl}
            else:
                names = {}
            trace_it(
                db,
                q.split(","),
                hex_to_bytes(contract),
                start,
                end,
                names,
                output_format,
            )
    else:
        print(
            f"Unsupported schema type {stype} for "
            f"currency {currency}. Only account is supported."
        )



import logging
from dataclasses import dataclass
from typing import Dict, List, Tuple

import pandas as pd
from cassandra.cqlengine.columns import Integer
from cassandra.cqlengine.usertype import UserType

from graphsenselib.deltaupdate.update.account.modelsdelta import (
    BalanceDelta,
    EntityDeltaAccount,
    RawEntityTxAccount,
    RelationDeltaAccount,
)
from graphsenselib.deltaupdate.update.account.modelsraw import Block, Trace, Transaction
from graphsenselib.deltaupdate.update.account.tokens import TokenTransfer
from graphsenselib.deltaupdate.update.generic import DeltaScalar, DeltaValue

logger = logging.getLogger(__name__)

currency_to_decimals = {
    "ETH": 18,
    "TRX": 6,
}


@dataclass
class TxReference(UserType):
    trace_index: Integer(required=False)
    log_index: Integer(required=False)


def only_call_traces(traces: List) -> List:
    return [trace for trace in traces if trace.call_type == "call"]


def get_prices(
    value, decimals, block_rates, usd_equivalent, coin_equivalent
) -> List[int]:
    euro_per_eth = block_rates[0]
    dollar_per_eth = block_rates[1]
    dollar_per_euro = dollar_per_eth / euro_per_eth

    if usd_equivalent == 1:
        dollar_value = value / 10**decimals
    elif coin_equivalent == 1:
        dollar_value = value / 10**decimals * dollar_per_eth
    else:
        raise Exception(
            "Unknown price type. only native coin and dollar equivalent supported atm"
        )

    euro_value = dollar_value / dollar_per_euro
    return [euro_value, dollar_value]


def get_prices_coin(value, currency, block_rates):
    coin_decimals = currency_to_decimals[currency]
    return get_prices(value, coin_decimals, block_rates, 0, 1)


def get_entitytx_from_tokentransfer(
    tokentransfer: TokenTransfer, is_outgoing, rates, hash_to_id, address_hash_to_id
) -> RawEntityTxAccount:
    tx_id = hash_to_id[tokentransfer.tx_hash]

    address_hash = (
        tokentransfer.from_address if is_outgoing else tokentransfer.to_address
    )
    address_id = address_hash_to_id[address_hash]

    dv = DeltaValue(
        tokentransfer.value,
        get_prices(
            tokentransfer.value,
            tokentransfer.decimals,
            rates[tokentransfer.block_id],
            tokentransfer.usd_equivalent,
            tokentransfer.coin_equivalent,
        ),
    )

    token_values = {tokentransfer.asset: dv}

    tx_reference = {
        "trace_index": None,
        "log_index": tokentransfer.log_index,
    }
    tx_reference = TxReference(**tx_reference)

    reta = RawEntityTxAccount(
        identifier=address_id,
        is_outgoing=is_outgoing,
        tx_id=tx_id,
        tx_reference=tx_reference,
        value=0,
        token_values=token_values,
        block_id=tokentransfer.block_id,
    )
    return reta


def get_entitytx_from_transaction(
    tx: Transaction, is_outgoing, hash_to_id, address_hash_to_id
) -> RawEntityTxAccount:
    tx_id = hash_to_id[tx.tx_hash]
    address_hash = tx.from_address if is_outgoing else tx.to_address

    address_id = address_hash_to_id[address_hash]

    tx_reference = {
        "trace_index": None,
        "log_index": None,
    }
    tx_reference = TxReference(**tx_reference)

    reta = RawEntityTxAccount(
        identifier=address_id,
        is_outgoing=is_outgoing,
        tx_id=tx_id,
        tx_reference=tx_reference,
        value=tx.value,
        token_values={},
        block_id=tx.block_id,
    )
    return reta


def balance_updates_traces_txs(
    relation_updates: List[RelationDeltaAccount],
    address_hash_to_id: Dict[bytes, int],
    currency: str,
) -> List[BalanceDelta]:
    excludedCallTypes = [
        "delegatecall",
        "callcode",
        "staticcall",
    ]
    filtered_updates = [
        update for update in relation_updates if update.type not in excludedCallTypes
    ]

    return [
        BalanceDelta(
            address_hash_to_id[update.src_identifier],
            {currency: DeltaScalar(-update.value.value)},
        )
        for update in filtered_updates
    ] + [
        BalanceDelta(
            address_hash_to_id[update.dst_identifier],
            {currency: DeltaScalar(update.value.value)},
        )
        for update in filtered_updates
        if update.dst_identifier is not None
    ]


def balance_updates_tokens(
    relation_updates: List[RelationDeltaAccount], address_hash_to_id: Dict[bytes, int]
) -> List[BalanceDelta]:
    updates = []
    for update in relation_updates:
        for token, value in update.token_values.items():
            updates.append(
                BalanceDelta(
                    address_hash_to_id[update.src_identifier],
                    {token: DeltaScalar(-value.value)},
                )
            )
            updates.append(
                BalanceDelta(
                    address_hash_to_id[update.dst_identifier],
                    {token: DeltaScalar(value.value)},
                )
            )

    return updates


def get_entitytx_from_trace(
    trace: Trace, is_outgoing: bool, hash_to_id: dict, address_hash_to_id: dict
) -> RawEntityTxAccount:
    tx_id = hash_to_id[trace.tx_hash]
    address_hash = trace.from_address if is_outgoing else trace.to_address
    address_id = address_hash_to_id[address_hash]

    tx_reference = {
        "trace_index": trace.trace_index,
        "log_index": None,
    }
    tx_reference = TxReference(**tx_reference)

    reta = RawEntityTxAccount(
        identifier=address_id,
        is_outgoing=is_outgoing,
        tx_id=tx_id,
        tx_reference=tx_reference,
        value=trace.value,
        token_values={},  # we dont support TRC10 right now
        block_id=trace.block_id,
    )
    return reta


def get_entitydelta_from_trace(
    trace: Trace,
    is_outgoing: bool,
    rates: Dict[int, Tuple[float, float]],
    hash_to_id: dict,
    currency: str,
) -> EntityDeltaAccount:
    identifier = trace.from_address if is_outgoing else trace.to_address
    total_received_value = 0 if is_outgoing else trace.value
    total_spent_value = trace.value if is_outgoing else 0
    total_received = DeltaValue(
        total_received_value,
        get_prices_coin(total_received_value, currency, rates[trace.block_id]),
    )
    total_spent = DeltaValue(
        total_spent_value,
        get_prices_coin(total_spent_value, currency, rates[trace.block_id]),
    )
    total_tokens_received = (
        {}
    )  # for now we dont support TRC10, so an empty dict is fine
    total_tokens_spent = {}  # for now we dont support TRC10, so an empty dict is fine
    if trace.tx_hash is None:
        first_tx_id = -1
        last_tx_id = -1
        no_incoming_txs = 0  # spark logic
    else:
        first_tx_id = hash_to_id[trace.tx_hash]
        last_tx_id = hash_to_id[trace.tx_hash]
        no_incoming_txs = int(not is_outgoing)

    no_outgoing_txs = int(is_outgoing)
    no_zerovalue = int((trace.value == 0))  # and trace.call_type == "call")
    no_incoming_txs_zero_value = 0 if is_outgoing else no_zerovalue
    no_outgoing_txs_zero_value = no_zerovalue if is_outgoing else 0

    eda = EntityDeltaAccount(
        identifier=identifier,
        total_received=total_received,
        total_spent=total_spent,
        total_tokens_received=total_tokens_received,
        total_tokens_spent=total_tokens_spent,
        first_tx_id=first_tx_id,
        last_tx_id=last_tx_id,
        no_incoming_txs=no_incoming_txs,
        no_outgoing_txs=no_outgoing_txs,
        no_incoming_txs_zero_value=no_incoming_txs_zero_value,
        no_outgoing_txs_zero_value=no_outgoing_txs_zero_value,
    )
    return eda


def get_entitydelta_from_tokentransfer(
    tokentransfer: TokenTransfer,
    is_outgoing: bool,
    rates: Dict[int, Tuple[float, float]],
    hash_to_id: dict,
) -> EntityDeltaAccount:
    identifier = tokentransfer.from_address if is_outgoing else tokentransfer.to_address

    fiat_values = get_prices(
        tokentransfer.value,
        tokentransfer.decimals,
        rates[tokentransfer.block_id],
        tokentransfer.usd_equivalent,
        tokentransfer.coin_equivalent,
    )
    dv = DeltaValue(tokentransfer.value, fiat_values)

    total_received = DeltaValue(0, [0, 0])
    total_spent = DeltaValue(0, [0, 0])
    total_tokens_received = {tokentransfer.asset: dv} if not is_outgoing else {}
    total_tokens_spent = {tokentransfer.asset: dv} if is_outgoing else {}
    first_tx_id = hash_to_id[tokentransfer.tx_hash]
    last_tx_id = hash_to_id[tokentransfer.tx_hash]
    no_incoming_txs = int(not is_outgoing)
    no_outgoing_txs = int(is_outgoing)
    no_incoming_txs_zero_value = 0
    no_outgoing_txs_zero_value = 0

    eda = EntityDeltaAccount(
        identifier=identifier,
        total_received=total_received,
        total_spent=total_spent,
        total_tokens_received=total_tokens_received,
        total_tokens_spent=total_tokens_spent,
        first_tx_id=first_tx_id,
        last_tx_id=last_tx_id,
        no_incoming_txs=no_incoming_txs,
        no_outgoing_txs=no_outgoing_txs,
        no_incoming_txs_zero_value=no_incoming_txs_zero_value,
        no_outgoing_txs_zero_value=no_outgoing_txs_zero_value,
    )
    return eda


def get_entitydelta_from_transaction(
    tx: Transaction,
    is_outgoing: bool,
    rates: Dict[int, Tuple[float, float]],
    hash_to_id: Dict[str, int],
    currency: str,
) -> EntityDeltaAccount:
    identifier = tx.from_address if is_outgoing else tx.to_address

    total_received_value = 0 if is_outgoing else tx.value
    total_spent_value = tx.value if is_outgoing else 0

    total_received = DeltaValue(
        total_received_value,
        get_prices_coin(total_received_value, currency, rates[tx.block_id]),
    )
    total_spent = DeltaValue(
        total_spent_value,
        get_prices_coin(total_spent_value, currency, rates[tx.block_id]),
    )
    total_tokens_received = {}
    total_tokens_spent = {}
    first_tx_id = hash_to_id[tx.tx_hash]
    last_tx_id = hash_to_id[tx.tx_hash]

    no_incoming_txs = int(not is_outgoing)
    no_outgoing_txs = int(is_outgoing)
    no_incoming_txs_zero_value = 0 if is_outgoing else int(tx.value == 0)
    no_outgoing_txs_zero_value = int(tx.value == 0) if is_outgoing else 0

    eda = EntityDeltaAccount(
        identifier=identifier,
        total_received=total_received,
        total_spent=total_spent,
        total_tokens_received=total_tokens_received,
        total_tokens_spent=total_tokens_spent,
        first_tx_id=first_tx_id,
        last_tx_id=last_tx_id,
        no_incoming_txs=no_incoming_txs,
        no_outgoing_txs=no_outgoing_txs,
        no_incoming_txs_zero_value=no_incoming_txs_zero_value,
        no_outgoing_txs_zero_value=no_outgoing_txs_zero_value,
    )
    return eda


def relationdelta_from_trace(
    trace: Trace, rates: Dict[int, Tuple[float, float]], currency: str
) -> RelationDeltaAccount:
    fadr, tadr = trace.from_address, trace.to_address
    value = DeltaValue(
        trace.value, get_prices_coin(trace.value, currency, rates[trace.block_id])
    )
    token_values = {}

    no_transactions = 1
    return RelationDeltaAccount(
        src_identifier=fadr,
        dst_identifier=tadr,
        no_transactions=no_transactions,
        value=value,
        token_values=token_values,
        type=trace.call_type,
    )


def relationdelta_from_transaction(
    tx: Transaction, rates: Dict[int, Tuple[float, float]], currency: str
) -> RelationDeltaAccount:
    iadr, oadr = tx.from_address, tx.to_address
    value = DeltaValue(
        tx.value, get_prices_coin(tx.value, currency, rates[tx.block_id])
    )
    token_values = {}
    no_transactions = 1

    return RelationDeltaAccount(
        src_identifier=iadr,
        dst_identifier=oadr,
        no_transactions=no_transactions,
        value=value,
        token_values=token_values,
        type="tx",
    )


def relationdelta_from_tokentransfer(
    tokentransfer: TokenTransfer, rates: Dict[int, Tuple[float, float]]
) -> RelationDeltaAccount:
    iadr, oadr = tokentransfer.from_address, tokentransfer.to_address
    value = tokentransfer.value
    dollar_value, euro_value = get_prices(
        tokentransfer.value,
        tokentransfer.decimals,
        rates[tokentransfer.block_id],
        tokentransfer.usd_equivalent,
        tokentransfer.coin_equivalent,
    )
    value = DeltaValue(value, [dollar_value, euro_value])

    token_values = {tokentransfer.asset: value}
    no_transactions = 1
    return RelationDeltaAccount(
        src_identifier=iadr,
        dst_identifier=oadr,
        no_transactions=no_transactions,
        value=DeltaValue(0, [0, 0]),
        token_values=token_values,
        type="token",
    )


def get_sorted_unique_addresses(
    traces_s: List[Trace],
    reward_traces: List[Trace],
    token_transfers: List[TokenTransfer],
    transactions: List[Transaction],
) -> pd.Series:
    addresses_sorting_df_to_tokens = [
        {
            "address": obj.to_address,
            "block_id": obj.block_id,
            "is_log": True,
            "index": obj.log_index,
            "is_from_address": False,
        }
        for obj in token_transfers
    ]

    addresses_sorting_df_from_tokens = [
        {
            "address": obj.from_address,
            "block_id": obj.block_id,
            "is_log": True,
            "index": obj.log_index,
            "is_from_address": True,
        }
        for obj in token_transfers
    ]

    addresses_sorting_df_to_traces = [
        {
            "address": obj.to_address,
            "block_id": obj.block_id,
            "is_log": False,
            "index": obj.trace_index,
            "is_from_address": False,
        }
        for obj in traces_s + reward_traces
    ]

    addresses_sorting_df_from_traces = [
        {
            "address": obj.from_address,
            "block_id": obj.block_id,
            "is_log": False,
            "index": obj.trace_index,
            "is_from_address": True,
        }
        for obj in traces_s
    ]

    addresses_sorting_df_from_txs = [
        {
            "address": obj.from_address,
            "block_id": obj.block_id,
            "is_log": False,
            # this is a hack to imitate spark; we assume there a max 1M tx per block
            "index": obj.transaction_index - 1_000_000,
            "is_from_address": True,
        }
        for obj in transactions
        if obj.from_address is not None
    ]
    addresses_sorting_df_to_txs = [
        {
            "address": obj.to_address,
            "block_id": obj.block_id,
            "is_log": False,
            # this is a hack to imitate spark; we assume there a max 1M tx per block
            "index": obj.transaction_index - 1_000_000,
            "is_from_address": False,
        }
        for obj in transactions
    ]

    addresses_sorting_df_data = (
        addresses_sorting_df_from_traces
        + addresses_sorting_df_to_traces
        + addresses_sorting_df_from_txs
        + addresses_sorting_df_to_txs
        + addresses_sorting_df_from_tokens
        + addresses_sorting_df_to_tokens
    )

    addresses_sorting_df = pd.DataFrame(addresses_sorting_df_data)
    if addresses_sorting_df.empty:
        return pd.Series([])

    addresses_sorting_df.sort_values(
        inplace=True, by=["block_id", "is_log", "index", "is_from_address"]
    )
    df_sorted_unique = addresses_sorting_df.drop_duplicates(
        keep="first", subset=["address"]
    )
    addresses = df_sorted_unique["address"]
    return addresses


def get_entity_transaction_updates_trace_token(
    traces_s_filtered: List[Trace],
    token_transfers: List[TokenTransfer],
    hash_to_id: dict,
    address_hash_to_id: dict,
    rates: dict,
) -> List[RawEntityTxAccount]:
    trace_outgoing = [
        get_entitytx_from_trace(trace, True, hash_to_id, address_hash_to_id)
        for trace in traces_s_filtered
    ]

    trace_incoming = [
        get_entitytx_from_trace(trace, False, hash_to_id, address_hash_to_id)
        for trace in traces_s_filtered
    ]

    token_outgoing = [
        get_entitytx_from_tokentransfer(tt, True, rates, hash_to_id, address_hash_to_id)
        for tt in token_transfers
    ]
    token_incoming = [
        get_entitytx_from_tokentransfer(
            tt, False, rates, hash_to_id, address_hash_to_id
        )
        for tt in token_transfers
    ]

    entity_transactions_traces_tokens = (
        trace_outgoing + trace_incoming + token_outgoing + token_incoming
    )
    return entity_transactions_traces_tokens


def get_entity_updates_trace_token(
    traces_s_filtered: List[Trace],
    token_transfers: List[TokenTransfer],
    reward_traces: List[Trace],
    hash_to_id: dict,
    currency: str,
    rates: dict,
):
    trace_outgoing = [
        get_entitydelta_from_trace(trace, True, rates, hash_to_id, currency)
        for trace in traces_s_filtered
    ]

    trace_incoming = [
        get_entitydelta_from_trace(trace, False, rates, hash_to_id, currency)
        for trace in traces_s_filtered + reward_traces
    ]

    token_outgoing = [
        get_entitydelta_from_tokentransfer(tt, True, rates, hash_to_id)
        for tt in token_transfers
    ]

    token_incoming = [
        get_entitydelta_from_tokentransfer(tt, False, rates, hash_to_id)
        for tt in token_transfers
    ]

    entity_deltas_traces_tokens = (
        trace_outgoing + trace_incoming + token_outgoing + token_incoming
    )
    return entity_deltas_traces_tokens


def get_entity_transactions_updates_tx(
    transactions: List[Transaction],
    hash_to_id: Dict[str, int],
    address_hash_to_id: Dict[bytes, int],
) -> List[RawEntityTxAccount]:
    outgoing = [
        get_entitytx_from_transaction(tx, True, hash_to_id, address_hash_to_id)
        for tx in transactions
        if tx.from_address is not None
    ]
    incoming = [
        get_entitytx_from_transaction(tx, False, hash_to_id, address_hash_to_id)
        for tx in transactions
        if tx.to_address is not None
    ]
    entity_transactions_tx = outgoing + incoming
    return entity_transactions_tx


def get_entity_updates_tx(
    transactions: List[Transaction],
    hash_to_id: Dict[str, int],
    currency: str,
    rates: Dict[int, Tuple[float, float]],
) -> List[EntityDeltaAccount]:
    outgoing = [
        get_entitydelta_from_transaction(tx, True, rates, hash_to_id, currency)
        for tx in transactions
        if tx.from_address is not None
    ]
    incoming = [
        get_entitydelta_from_transaction(tx, False, rates, hash_to_id, currency)
        for tx in transactions
        if tx.to_address is not None
    ]

    entity_deltas_tx = outgoing + incoming

    return entity_deltas_tx


def get_balance_deltas(
    relation_updates_trace: List[RelationDeltaAccount],
    relation_updates_tx: List[RelationDeltaAccount],
    relation_updates_tokens: List[RelationDeltaAccount],
    reward_traces: List[Trace],
    transactions: List[Transaction],
    blocks: List[Block],
    address_hash_to_id: Dict[bytes, int],
    currency: str,
) -> List[BalanceDelta]:
    credits_debits_tokens_eth = []
    credits_debits_tokens_eth += balance_updates_traces_txs(
        relation_updates_trace + relation_updates_tx, address_hash_to_id, currency
    )
    credits_debits_tokens_eth += balance_updates_tokens(
        relation_updates_tokens, address_hash_to_id
    )

    miner_rewards = [
        BalanceDelta(address_hash_to_id[t.to_address], {currency: DeltaScalar(t.value)})
        for t in reward_traces
    ]

    if currency == "TRX":
        txFeeDebits = []
        burntFees = []
        txFeeCredits = [
            BalanceDelta(
                address_hash_to_id[tx.from_address],
                {currency: DeltaScalar(-tx.fee)},
            )
            for tx in transactions
            if tx.from_address in address_hash_to_id
        ]
    elif currency == "ETH":
        block_to_miner_id = {
            block.block_id: address_hash_to_id[block.miner] for block in blocks
        }
        txFeeDebits = [
            BalanceDelta(
                block_to_miner_id[tx.block_id],
                {currency: DeltaScalar(tx.receipt_gas_used * tx.gas_price)},
            )
            for tx in transactions
        ]

        burntFees = [
            BalanceDelta(
                block_to_miner_id[b.block_id],
                {currency: DeltaScalar(-b.base_fee_per_gas * b.gas_used)},
            )
            for b in blocks
        ]
        txFeeCredits = [
            BalanceDelta(
                address_hash_to_id[tx.from_address],
                {currency: DeltaScalar(-tx.receipt_gas_used * tx.gas_price)},
            )
            for tx in transactions
            if tx.from_address in address_hash_to_id
        ]
    else:
        raise ValueError(f"Unknown currency {currency}")

    balance_updates = (
        credits_debits_tokens_eth
        + txFeeDebits
        + txFeeCredits
        + burntFees
        + miner_rewards
    )
    return balance_updates
